# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest

from ansys.fluent.core import examples


@pytest.mark.fluent_version(">=26.1")
def test_new_watertight_workflow(new_meshing_session_wo_exit):
    # Import geometry
    import_file_name = examples.download_file(
        "mixing_elbow.pmdb", "pyfluent/mixing_elbow"
    )
    new_meshing_session_wo_exit.workflow.InitializeWorkflow(
        WorkflowType="Watertight Geometry"
    )
    watertight = new_meshing_session_wo_exit.meshing_workflow
    watertight.task_object.import_geometry["Import Geometry"].arguments.file_name = (
        import_file_name
    )
    assert (
        watertight.task_object.import_geometry[
            "Import Geometry"
        ].arguments.length_unit()
        == "mm"
    )
    watertight.task_object.import_geometry[
        "Import Geometry"
    ].arguments.length_unit.set_state("in")
    assert (
        watertight.task_object.import_geometry[
            "Import Geometry"
        ].arguments.length_unit.get_state()
        == "in"
    )
    watertight.task_object.import_geometry["Import Geometry"].execute()

    # Add local sizing
    watertight.task_object.add_local_sizing_wtm["Add Local Sizing"].add_child_to_task()
    watertight.task_object.add_local_sizing_wtm["Add Local Sizing"].execute()

    # Generate surface mesh
    watertight.task_object.create_surface_mesh[
        "Generate the Surface Mesh"
    ].arguments.cfd_surface_mesh_controls.max_size = 0.3
    assert (
        watertight.task_object.create_surface_mesh[
            "Generate the Surface Mesh"
        ].arguments.cfd_surface_mesh_controls.max_size()
        == 0.3
    )
    watertight.task_object.create_surface_mesh["Generate the Surface Mesh"].execute()

    # Describe geometry
    watertight.task_object.describe_geometry["Describe Geometry"].update_child_tasks(
        setup_type_changed=False
    )
    watertight.task_object.describe_geometry[
        "Describe Geometry"
    ].arguments.setup_type.set_state(
        "The geometry consists of only fluid regions with no voids"
    )
    watertight.task_object.describe_geometry["Describe Geometry"].update_child_tasks(
        setup_type_changed=False
    )
    watertight.task_object.describe_geometry["Describe Geometry"].execute()

    # Update boundaries
    watertight.task_object.update_boundaries[
        "Update Boundaries"
    ].arguments.boundary_label_list.set_state(["wall-inlet"])
    watertight.task_object.update_boundaries[
        "Update Boundaries"
    ].arguments.boundary_label_type_list.set_state(["wall"])
    watertight.task_object.update_boundaries[
        "Update Boundaries"
    ].arguments.old_boundary_label_list.set_state(["wall-inlet"])
    watertight.task_object.update_boundaries[
        "Update Boundaries"
    ].arguments.old_boundary_label_type_list.set_state(["velocity-inlet"])
    watertight.task_object.update_boundaries["Update Boundaries"].execute()

    # Update regions
    watertight.task_object.update_regions["Update Regions"].execute()

    # Add boundary layers
    watertight.task_object.add_boundary_layers[
        "Add Boundary Layers"
    ].add_child_to_task()
    watertight.task_object.add_boundary_layers[
        "Add Boundary Layers"
    ].arguments.control_name.set_state("smooth-transition_1")
    watertight.task_object.add_boundary_layers[
        "Add Boundary Layers"
    ].insert_compound_child_task()
    watertight.task_object.add_boundary_layers["Add Boundary Layers"].execute()

    # Generate volume mesh
    watertight.task_object.create_volume_mesh_wtm[
        "Generate the Volume Mesh"
    ].arguments.volume_fill.set_state("poly-hexcore")
    watertight.task_object.create_volume_mesh_wtm[
        "Generate the Volume Mesh"
    ].arguments.volume_fill_controls.hex_max_cell_length = 0.3
    watertight.task_object.create_volume_mesh_wtm["Generate the Volume Mesh"].execute()

    # Switch to solution mode
    solver = new_meshing_session_wo_exit.switch_to_solver()
    assert solver.is_active() is True
    assert new_meshing_session_wo_exit.is_active() is False
    solver.exit()
    assert solver.is_active() is False


@pytest.mark.fluent_version(">=26.1")
def test_new_fault_tolerant_workflow(new_meshing_session_wo_exit):
    meshing = new_meshing_session_wo_exit

    # Import CAD and part management
    import_file_name = examples.download_file(
        "exhaust_system.fmd", "pyfluent/exhaust_system"
    )
    new_meshing_session_wo_exit.workflow.InitializeWorkflow(
        WorkflowType="Fault-tolerant Meshing"
    )
    fault_tolerant = meshing.meshing_workflow
    meshing.PartManagement.InputFileChanged(
        FilePath=import_file_name, IgnoreSolidNames=False, PartPerBody=False
    )
    meshing.PMFileManagement.FileManager.LoadFiles()
    meshing.PartManagement.Node["Meshing Model"].Copy(
        Paths=[
            "/dirty_manifold-for-wrapper," + "1/dirty_manifold-for-wrapper,1/main,1",
            "/dirty_manifold-for-wrapper,"
            + "1/dirty_manifold-for-wrapper,1/flow-pipe,1",
            "/dirty_manifold-for-wrapper,"
            + "1/dirty_manifold-for-wrapper,1/outpipe3,1",
            "/dirty_manifold-for-wrapper," + "1/dirty_manifold-for-wrapper,1/object2,1",
            "/dirty_manifold-for-wrapper," + "1/dirty_manifold-for-wrapper,1/object1,1",
        ]
    )
    meshing.PartManagement.ObjectSetting["DefaultObjectSetting"].OneZonePer.set_state(
        "part"
    )
    fault_tolerant.task_object.import_cad_and_part_management[
        "Import CAD and Part Management"
    ].arguments.context.set_state(0)
    fault_tolerant.task_object.import_cad_and_part_management[
        "Import CAD and Part Management"
    ].arguments.create_object_per = "Custom"
    fault_tolerant.task_object.import_cad_and_part_management[
        "Import CAD and Part Management"
    ].arguments.fmd_file_name.set_state(import_file_name)
    fault_tolerant.task_object.import_cad_and_part_management[
        "Import CAD and Part Management"
    ].arguments.file_loaded = True
    fault_tolerant.task_object.import_cad_and_part_management[
        "Import CAD and Part Management"
    ].arguments.object_setting = "DefaultObjectSetting"
    fault_tolerant.task_object.import_cad_and_part_management[
        "Import CAD and Part Management"
    ].execute()

    # Describe geometry and flow
    fault_tolerant.task_object.describe_geometry_and_flow[
        "Describe Geometry and Flow"
    ].arguments.add_enclosure.set_state(False)
    fault_tolerant.task_object.describe_geometry_and_flow[
        "Describe Geometry and Flow"
    ].arguments.close_caps = True
    fault_tolerant.task_object.describe_geometry_and_flow[
        "Describe Geometry and Flow"
    ].arguments.describe_geometry_and_flow_options.advanced_options = True
    fault_tolerant.task_object.describe_geometry_and_flow[
        "Describe Geometry and Flow"
    ].arguments.describe_geometry_and_flow_options.extract_edge_features = True
    fault_tolerant.task_object.describe_geometry_and_flow[
        "Describe Geometry and Flow"
    ].arguments.flow_type = "Internal flow through the object"
    fault_tolerant.task_object.describe_geometry_and_flow[
        "Describe Geometry and Flow"
    ].update_child_tasks(setup_type_changed=False)
    fault_tolerant.task_object.describe_geometry_and_flow[
        "Describe Geometry and Flow"
    ].execute()

    # Enclose fluid regions (capping)
    fault_tolerant.task_object.capping[
        "Enclose Fluid Regions (Capping)"
    ].arguments.patch_name.set_state("inlet-1")
    fault_tolerant.task_object.capping[
        "Enclose Fluid Regions (Capping)"
    ].arguments.selection_type.set_state("zone")
    fault_tolerant.task_object.capping[
        "Enclose Fluid Regions (Capping)"
    ].arguments.zone_selection_list.set_state(["inlet.1"])
    fault_tolerant.task_object.capping[
        "Enclose Fluid Regions (Capping)"
    ].insert_compound_child_task()
    fault_tolerant.task_object.capping["inlet-1"].execute()

    fault_tolerant.task_object.capping[
        "Enclose Fluid Regions (Capping)"
    ].arguments.patch_name.set_state("inlet-2")
    fault_tolerant.task_object.capping[
        "Enclose Fluid Regions (Capping)"
    ].arguments.selection_type.set_state("zone")
    fault_tolerant.task_object.capping[
        "Enclose Fluid Regions (Capping)"
    ].arguments.zone_selection_list.set_state(["inlet.2"])
    fault_tolerant.task_object.capping[
        "Enclose Fluid Regions (Capping)"
    ].insert_compound_child_task()
    fault_tolerant.task_object.capping["inlet-2"].execute()

    fault_tolerant.task_object.capping[
        "Enclose Fluid Regions (Capping)"
    ].arguments.patch_name.set_state("inlet-3")
    fault_tolerant.task_object.capping[
        "Enclose Fluid Regions (Capping)"
    ].arguments.selection_type.set_state("zone")
    fault_tolerant.task_object.capping[
        "Enclose Fluid Regions (Capping)"
    ].arguments.zone_selection_list.set_state(["inlet"])
    fault_tolerant.task_object.capping[
        "Enclose Fluid Regions (Capping)"
    ].insert_compound_child_task()
    fault_tolerant.task_object.capping["inlet-3"].execute()

    fault_tolerant.task_object.capping[
        "Enclose Fluid Regions (Capping)"
    ].arguments.patch_name.set_state("outlet-1")
    fault_tolerant.task_object.capping[
        "Enclose Fluid Regions (Capping)"
    ].arguments.selection_type.set_state("zone")
    fault_tolerant.task_object.capping[
        "Enclose Fluid Regions (Capping)"
    ].arguments.zone_selection_list.set_state(["outlet"])
    fault_tolerant.task_object.capping[
        "Enclose Fluid Regions (Capping)"
    ].arguments.zone_type.set_state("pressure-outlet")
    fault_tolerant.task_object.capping[
        "Enclose Fluid Regions (Capping)"
    ].insert_compound_child_task()
    fault_tolerant.task_object.capping["outlet-1"].execute()

    # Extract edge features
    fault_tolerant.task_object.extract_edge_features[
        "Extract Edge Features"
    ].arguments.extract_edges_name.set_state("edge-group-1")
    fault_tolerant.task_object.extract_edge_features[
        "Extract Edge Features"
    ].arguments.extract_method_type.set_state("Intersection Loops")
    fault_tolerant.task_object.extract_edge_features[
        "Extract Edge Features"
    ].arguments.object_selection_list.set_state(["flow_pipe", "main"])
    fault_tolerant.task_object.extract_edge_features[
        "Extract Edge Features"
    ].insert_compound_child_task()
    fault_tolerant.task_object.extract_edge_features["edge-group-1"].execute()

    # Identify regions
    fault_tolerant.task_object.identify_regions[
        "Identify Regions"
    ].arguments.show_coordinates = True
    fault_tolerant.task_object.identify_regions[
        "Identify Regions"
    ].arguments.material_points_name.set_state("fluid-region-1")
    fault_tolerant.task_object.identify_regions[
        "Identify Regions"
    ].arguments.selection_type.set_state("zone")
    fault_tolerant.task_object.identify_regions[
        "Identify Regions"
    ].arguments.x.set_state(377.322045740589)
    fault_tolerant.task_object.identify_regions[
        "Identify Regions"
    ].arguments.y.set_state(-176.800676988458)
    fault_tolerant.task_object.identify_regions[
        "Identify Regions"
    ].arguments.z.set_state(-37.0764628583475)
    fault_tolerant.task_object.identify_regions[
        "Identify Regions"
    ].arguments.zone_selection_list.set_state(["main.1"])
    fault_tolerant.task_object.identify_regions[
        "Identify Regions"
    ].insert_compound_child_task()
    fault_tolerant.task_object.identify_regions["fluid-region-1"].execute()

    fault_tolerant.task_object.identify_regions[
        "Identify Regions"
    ].arguments.show_coordinates = True
    fault_tolerant.task_object.identify_regions[
        "Identify Regions"
    ].arguments.material_points_name.set_state("void-region-1")
    fault_tolerant.task_object.identify_regions[
        "Identify Regions"
    ].arguments.new_region_type.set_state("void")
    fault_tolerant.task_object.identify_regions[
        "Identify Regions"
    ].arguments.object_selection_list.set_state(
        ["inlet-1", "inlet-2", "inlet-3", "main"]
    )
    fault_tolerant.task_object.identify_regions[
        "Identify Regions"
    ].arguments.x.set_state(374.722045740589)
    fault_tolerant.task_object.identify_regions[
        "Identify Regions"
    ].arguments.y.set_state(-278.9775145640143)
    fault_tolerant.task_object.identify_regions[
        "Identify Regions"
    ].arguments.z.set_state(-161.1700719416913)
    fault_tolerant.task_object.identify_regions[
        "Identify Regions"
    ].arguments.zone_selection_list.set_state(["main.1"])
    fault_tolerant.task_object.identify_regions[
        "Identify Regions"
    ].insert_compound_child_task()
    fault_tolerant.task_object.identify_regions["void-region-1"].execute()

    # Define leakage threshold
    fault_tolerant.task_object.define_leakage_threshold[
        "Define Leakage Threshold"
    ].arguments.set_state(
        {
            "add_child": "yes",
            "flip_direction": True,
            "leakage_name": "leakage-1",
            "plane_direction": "X",
            "region_selection_single": "void-region-1",
        }
    )
    fault_tolerant.task_object.define_leakage_threshold[
        "Define Leakage Threshold"
    ].insert_compound_child_task()
    fault_tolerant.task_object.define_leakage_threshold["leakage-1"].execute()

    # Update regions settings
    fault_tolerant.task_object.update_region_settings[
        "Update Region Settings"
    ].arguments.all_region_filter_categories.set_state(["2"] * 5 + ["1"] * 2)
    fault_tolerant.task_object.update_region_settings[
        "Update Region Settings"
    ].arguments.all_region_leakage_size_list.set_state(["none"] * 6 + ["6.4"])
    fault_tolerant.task_object.update_region_settings[
        "Update Region Settings"
    ].arguments.all_region_linked_construction_surface_list.set_state(
        ["n/a"] * 6 + ["no"]
    )
    fault_tolerant.task_object.update_region_settings[
        "Update Region Settings"
    ].arguments.all_region_mesh_method_list.set_state(["none"] * 6 + ["wrap"])
    fault_tolerant.task_object.update_region_settings[
        "Update Region Settings"
    ].arguments.all_region_name_list.set_state(
        [
            "main",
            "flow_pipe",
            "outpipe3",
            "object2",
            "object1",
            "void-region-1",
            "fluid-region-1",
        ]
    )
    fault_tolerant.task_object.update_region_settings[
        "Update Region Settings"
    ].arguments.all_region_overset_componen_list.set_state(["no"] * 7)
    fault_tolerant.task_object.update_region_settings[
        "Update Region Settings"
    ].arguments.all_region_source_list.set_state(["object"] * 5 + ["mpt"] * 2)
    fault_tolerant.task_object.update_region_settings[
        "Update Region Settings"
    ].arguments.all_region_type_list.set_state(["void"] * 6 + ["fluid"])
    fault_tolerant.task_object.update_region_settings[
        "Update Region Settings"
    ].arguments.all_region_volume_fill_list.set_state(["none"] * 6 + ["tet"])
    fault_tolerant.task_object.update_region_settings[
        "Update Region Settings"
    ].arguments.filter_category.set_state("Identified Regions")
    fault_tolerant.task_object.update_region_settings[
        "Update Region Settings"
    ].arguments.all_region_leakage_size_list.set_state([""])
    fault_tolerant.task_object.update_region_settings[
        "Update Region Settings"
    ].arguments.all_region_mesh_method_list.set_state(["wrap"])
    fault_tolerant.task_object.update_region_settings[
        "Update Region Settings"
    ].arguments.all_region_name_list.set_state(["fluid-region-1"])
    fault_tolerant.task_object.update_region_settings[
        "Update Region Settings"
    ].arguments.all_region_overset_componen_list.set_state(["no"])
    fault_tolerant.task_object.update_region_settings[
        "Update Region Settings"
    ].arguments.all_region_type_list.set_state(["fluid"])
    fault_tolerant.task_object.update_region_settings[
        "Update Region Settings"
    ].arguments.all_region_volume_fill_list.set_state(["hexcore"])
    fault_tolerant.task_object.update_region_settings[
        "Update Region Settings"
    ].arguments.all_region_leakage_size_list.set_state([""])
    fault_tolerant.task_object.update_region_settings[
        "Update Region Settings"
    ].arguments.all_region_mesh_method_list.set_state(["wrap"])
    fault_tolerant.task_object.update_region_settings[
        "Update Region Settings"
    ].arguments.all_region_name_list.set_state(["fluid-region-1"])
    fault_tolerant.task_object.update_region_settings[
        "Update Region Settings"
    ].arguments.all_region_overset_componen_list.set_state(["no"])
    fault_tolerant.task_object.update_region_settings[
        "Update Region Settings"
    ].arguments.all_region_type_list.set_state(["fluid"])
    fault_tolerant.task_object.update_region_settings[
        "Update Region Settings"
    ].arguments.all_region_volume_fill_list.set_state(["tet"])
    fault_tolerant.task_object.update_region_settings[
        "Update Region Settings"
    ].execute()

    # Choose mesh control options
    fault_tolerant.task_object.choose_mesh_control_options[
        "Choose Mesh Control Options"
    ].execute()

    # Generate surface mesh
    fault_tolerant.task_object.generate_surface_mesh[
        "Generate the Surface Mesh"
    ].execute()

    # Update boundaries
    fault_tolerant.task_object.update_boundaries["Update Boundaries"].execute()

    # Add boundary layers
    fault_tolerant.task_object.add_boundary_layers[
        "Add Boundary Layers"
    ].arguments.control_name.set_state("aspect-ratio_1")
    fault_tolerant.task_object.add_boundary_layers[
        "Add Boundary Layers"
    ].insert_compound_child_task()
    fault_tolerant.task_object.add_boundary_layers["aspect-ratio_1"].execute()

    # Generate volume mesh
    fault_tolerant.task_object.create_volume_mesh_ftm[
        "Generate the Volume Mesh"
    ].arguments.all_region_name_list.set_state(
        [
            "main",
            "flow_pipe",
            "outpipe3",
            "object2",
            "object1",
            "void-region-1",
            "fluid-region-1",
        ]
    )
    fault_tolerant.task_object.create_volume_mesh_ftm[
        "Generate the Volume Mesh"
    ].arguments.all_region_size_list.set_state(["11.33375"] * 7)
    fault_tolerant.task_object.create_volume_mesh_ftm[
        "Generate the Volume Mesh"
    ].arguments.all_region_volume_fill_list.set_state(["none"] * 6 + ["tet"])
    fault_tolerant.task_object.create_volume_mesh_ftm[
        "Generate the Volume Mesh"
    ].execute()

    # Generate volume mesh
    solver = meshing.switch_to_solver()
    assert solver.is_active() is True
    assert meshing.is_active() is False
    solver.exit()
    assert solver.is_active() is False


@pytest.mark.fluent_version(">=26.1")
def test_new_2d_meshing_workflow(new_meshing_session_wo_exit):
    # Import geometry
    import_file_name = examples.download_file("NACA0012.fmd", "pyfluent/airfoils")
    new_meshing_session_wo_exit.workflow.InitializeWorkflow(WorkflowType="2D Meshing")

    two_dim_mesh = new_meshing_session_wo_exit.meshing_workflow

    two_dim_mesh.task_object.load_cad_geometry[
        "Load CAD Geometry"
    ].arguments.file_name = import_file_name
    two_dim_mesh.task_object.load_cad_geometry[
        "Load CAD Geometry"
    ].arguments.length_unit = "mm"
    two_dim_mesh.task_object.load_cad_geometry[
        "Load CAD Geometry"
    ].arguments.refaceting.refacet = False
    two_dim_mesh.task_object.load_cad_geometry["Load CAD Geometry"].execute()

    # Set regions and boundaries
    two_dim_mesh.task_object.update_boundaries[
        "Update Boundaries"
    ].arguments.selection_type = "zone"
    two_dim_mesh.task_object.update_boundaries["Update Boundaries"].execute()

    # Define global sizing
    two_dim_mesh.task_object.define_global_sizing[
        "Define Global Sizing"
    ].arguments.curvature_normal_angle = 20
    two_dim_mesh.task_object.define_global_sizing[
        "Define Global Sizing"
    ].arguments.max_size = 2000.0
    two_dim_mesh.task_object.define_global_sizing[
        "Define Global Sizing"
    ].arguments.min_size = 5.0
    two_dim_mesh.task_object.define_global_sizing[
        "Define Global Sizing"
    ].arguments.size_functions = "Curvature"
    two_dim_mesh.task_object.define_global_sizing["Define Global Sizing"].execute()

    # Add local sizing
    two_dim_mesh.task_object.add_local_sizing_wtm[
        "Add Local Sizing"
    ].arguments.add_child = True
    two_dim_mesh.task_object.add_local_sizing_wtm[
        "Add Local Sizing"
    ].arguments.boi_control_name = "boi_1"
    two_dim_mesh.task_object.add_local_sizing_wtm[
        "Add Local Sizing"
    ].arguments.boi_execution = "Body Of Influence"
    two_dim_mesh.task_object.add_local_sizing_wtm[
        "Add Local Sizing"
    ].arguments.boi_face_label_list = ["boi"]
    two_dim_mesh.task_object.add_local_sizing_wtm[
        "Add Local Sizing"
    ].arguments.boi_size = 50.0
    two_dim_mesh.task_object.add_local_sizing_wtm[
        "Add Local Sizing"
    ].arguments.boi_zoneor_label = "label"
    two_dim_mesh.task_object.add_local_sizing_wtm[
        "Add Local Sizing"
    ].arguments.draw_size_control = True
    two_dim_mesh.task_object.add_local_sizing_wtm[
        "Add Local Sizing"
    ].add_child_and_update(defer_update=False)

    two_dim_mesh.task_object.add_local_sizing_wtm[
        "Add Local Sizing"
    ].arguments.add_child = True
    two_dim_mesh.task_object.add_local_sizing_wtm[
        "Add Local Sizing"
    ].arguments.boi_control_name = "edgesize_1"
    two_dim_mesh.task_object.add_local_sizing_wtm[
        "Add Local Sizing"
    ].arguments.boi_execution = "Edge Size"
    two_dim_mesh.task_object.add_local_sizing_wtm[
        "Add Local Sizing"
    ].arguments.boi_size = 5.0
    two_dim_mesh.task_object.add_local_sizing_wtm[
        "Add Local Sizing"
    ].arguments.boi_zoneor_label = "label"
    two_dim_mesh.task_object.add_local_sizing_wtm[
        "Add Local Sizing"
    ].arguments.draw_size_control = True
    two_dim_mesh.task_object.add_local_sizing_wtm[
        "Add Local Sizing"
    ].arguments.edge_label_list = ["airfoil-te"]
    two_dim_mesh.task_object.add_local_sizing_wtm[
        "Add Local Sizing"
    ].add_child_and_update(defer_update=False)

    two_dim_mesh.task_object.add_local_sizing_wtm[
        "Add Local Sizing"
    ].arguments.add_child = True
    two_dim_mesh.task_object.add_local_sizing_wtm[
        "Add Local Sizing"
    ].arguments.boi_control_name = "curvature_1"
    two_dim_mesh.task_object.add_local_sizing_wtm[
        "Add Local Sizing"
    ].arguments.boi_curvature_normal_angle = 10
    two_dim_mesh.task_object.add_local_sizing_wtm[
        "Add Local Sizing"
    ].arguments.boi_execution = "Curvature"
    two_dim_mesh.task_object.add_local_sizing_wtm[
        "Add Local Sizing"
    ].arguments.boi_max_size = 2
    two_dim_mesh.task_object.add_local_sizing_wtm[
        "Add Local Sizing"
    ].arguments.boi_min_size = 1.5
    two_dim_mesh.task_object.add_local_sizing_wtm[
        "Add Local Sizing"
    ].arguments.boi_scope_to = "edges"
    two_dim_mesh.task_object.add_local_sizing_wtm[
        "Add Local Sizing"
    ].arguments.boi_zoneor_label = "label"
    two_dim_mesh.task_object.add_local_sizing_wtm[
        "Add Local Sizing"
    ].arguments.draw_size_control = True
    two_dim_mesh.task_object.add_local_sizing_wtm[
        "Add Local Sizing"
    ].arguments.edge_label_list = ["airfoil"]
    two_dim_mesh.task_object.add_local_sizing_wtm[
        "Add Local Sizing"
    ].add_child_and_update(defer_update=False)

    # Add boundary layer
    two_dim_mesh.task_object.add_2d_boundary_layers[
        "Add 2D Boundary Layers"
    ].arguments.add_child = True
    two_dim_mesh.task_object.add_2d_boundary_layers[
        "Add 2D Boundary Layers"
    ].arguments.control_name = "aspect-ratio_1"
    two_dim_mesh.task_object.add_2d_boundary_layers[
        "Add 2D Boundary Layers"
    ].arguments.number_of_layers = 4
    two_dim_mesh.task_object.add_2d_boundary_layers[
        "Add 2D Boundary Layers"
    ].arguments.offset_method_type = "aspect-ratio"
    two_dim_mesh.task_object.add_2d_boundary_layers[
        "Add 2D Boundary Layers"
    ].add_child_and_update(defer_update=False)

    # NOTE: Setting `show_advanced_options = True` is required to configure advanced preferences.
    # This dependency may be removed in a future release as the API evolves.
    two_dim_mesh.task_object.generate_initial_surface_mesh[
        "Generate the Surface Mesh"
    ].arguments.surface_2d_preferences.show_advanced_options = True
    two_dim_mesh.task_object.generate_initial_surface_mesh[
        "Generate the Surface Mesh"
    ].arguments.surface_2d_preferences.merge_edge_zones_based_on_labels = False
    two_dim_mesh.task_object.generate_initial_surface_mesh[
        "Generate the Surface Mesh"
    ].arguments.surface_2d_preferences.merge_face_zones_based_on_labels = False
    two_dim_mesh.task_object.generate_initial_surface_mesh[
        "Generate the Surface Mesh"
    ].execute()

    two_dim_mesh.task_object.add_2d_boundary_layers["Add 2D Boundary Layers"].revert()
    two_dim_mesh.task_object.add_2d_boundary_layers[
        "Add 2D Boundary Layers"
    ].arguments.add_child = "yes"
    two_dim_mesh.task_object.add_2d_boundary_layers[
        "Add 2D Boundary Layers"
    ].arguments.control_name = "uniform_1"
    two_dim_mesh.task_object.add_2d_boundary_layers[
        "Add 2D Boundary Layers"
    ].arguments.first_layer_height = 2
    two_dim_mesh.task_object.add_2d_boundary_layers[
        "Add 2D Boundary Layers"
    ].arguments.number_of_layers = 4
    two_dim_mesh.task_object.add_2d_boundary_layers[
        "Add 2D Boundary Layers"
    ].arguments.offset_method_type = "uniform"
    two_dim_mesh.task_object.add_2d_boundary_layers["Add 2D Boundary Layers"].execute()

    # NOTE: Setting `show_advanced_options = True` is required to configure advanced preferences.
    # This dependency may be removed in a future release as the API evolves.
    two_dim_mesh.task_object.generate_initial_surface_mesh[
        "Generate the Surface Mesh"
    ].arguments.surface_2d_preferences.show_advanced_options = True
    two_dim_mesh.task_object.generate_initial_surface_mesh[
        "Generate the Surface Mesh"
    ].arguments.surface_2d_preferences.merge_edge_zones_based_on_labels = False
    two_dim_mesh.task_object.generate_initial_surface_mesh[
        "Generate the Surface Mesh"
    ].arguments.surface_2d_preferences.merge_face_zones_based_on_labels = False
    two_dim_mesh.task_object.generate_initial_surface_mesh[
        "Generate the Surface Mesh"
    ].execute()

    # Switch to solution mode
    solver = new_meshing_session_wo_exit.switch_to_solver()
    assert solver.is_active() is True
    assert new_meshing_session_wo_exit.is_active() is False
    solver.exit()
    assert solver.is_active() is False


@pytest.mark.fluent_version(">=26.1")
def test_arguments_and_parameters_in_new_meshing_workflow(new_meshing_session):
    new_meshing_session.workflow.InitializeWorkflow(WorkflowType="Watertight Geometry")
    watertight = new_meshing_session.meshing_workflow

    # arguments
    assert (
        watertight.task_object.import_geometry[
            "Import Geometry"
        ].arguments.length_unit()
        == "mm"
    )
    assert watertight.task_object.import_geometry[
        "Import Geometry"
    ].arguments.length_unit.allowed_values() == [
        "m",
        "cm",
        "mm",
        "in",
        "ft",
        "um",
        "nm",
    ]
    watertight.task_object.import_geometry["Import Geometry"].arguments.length_unit = (
        "m"
    )
    assert (
        watertight.task_object.import_geometry[
            "Import Geometry"
        ].arguments.length_unit.get_state()
        == "m"
    )
    watertight.task_object.import_geometry["Import Geometry"].arguments.set_state(
        {"length_unit": "in"}
    )
    assert (
        watertight.task_object.import_geometry[
            "Import Geometry"
        ].arguments.length_unit()
        == "in"
    )
    watertight.task_object.import_geometry["Import Geometry"].set_state(
        {"arguments": {"length_unit": "ft"}}
    )
    assert (
        watertight.task_object.import_geometry[
            "Import Geometry"
        ].arguments.length_unit()
        == "ft"
    )

    # parameters
    assert (
        watertight.task_object.import_geometry["Import Geometry"].state()
        == "Out-of-date"
    )
    assert watertight.task_object.import_geometry[
        "Import Geometry"
    ].state.allowed_values() == [
        "Out-of-date",
        "Attention-required",
        "Up-to-date",
        "Forced-up-to-date",
    ]
    watertight.task_object.import_geometry["Import Geometry"].state = "Up-to-date"
    assert (
        watertight.task_object.import_geometry["Import Geometry"].state.get_state()
        == "Up-to-date"
    )
    watertight.task_object.import_geometry["Import Geometry"].set_state(
        {"state": "Out-of-date"}
    )
    assert (
        watertight.task_object.import_geometry["Import Geometry"].state()
        == "Out-of-date"
    )
    assert (
        watertight.task_object.import_geometry[
            "Import Geometry"
        ].check_point.default_value()
        == "default-off"
    )

    # Both in a single set_state operation
    watertight.task_object.import_geometry["Import Geometry"].set_state(
        {"arguments": {"length_unit": "cm"}, "state": "Forced-up-to-date"}
    )
    assert (
        watertight.task_object.import_geometry[
            "Import Geometry"
        ].arguments.length_unit()
        == "cm"
    )
    assert (
        watertight.task_object.import_geometry["Import Geometry"].state()
        == "Forced-up-to-date"
    )



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

""".. _battery_pack:

Simulating a 1P3S Battery Pack Using the Battery Model
------------------------------------------------------
"""

# %%
# Objective
# ---------
#
# This example demonstrates a complete 1P3S battery pack simulation workflow comprising
# one parallel string and three cells in series using the NTGK
# (Newman-Tiedemann-Gu-Kim) electrochemical model utilizing PyFluent APIs.
# The goal is to model and analyze battery pack behavior under a constant
# 200 W discharge condition. This includes setting up active and passive zones,
# defining dual-path electrical conductivity using User-Defined Scalars (UDS),
# and applying convection cooling to simulate heat dissipation. The simulation
# aims to monitor key performance indicators such as pack voltage,
# maximum temperature, current flow, and heat generation over a 1500 second
# transient run, providing insights into thermal and electrical performance
# relevant to electric vehicle (EV) and energy storage system (ESS) design.

# %%
# Problem Description
# -------------------
#
# Simulation of a 1P3S lithium-ion battery pack, consisting of three cells
# connected in series and no parallel branches, undergoing a constant 200 W
# discharge to represent a high-load condition. Each cell has a
# nominal capacity of 14.6 Ah, resulting in a total pack capacity of 14.6 Ah
# and a nominal voltage range of 10.8–12.6 V, corresponding to an average cell
# voltage of approximately 3.9 V. The discharge C-rate equates to about 1.17C,
# and the total energy of roughly 171 Wh means the discharge would last around
# 51 minutes (3060 seconds). The simulation model includes active zones for the
# electrochemical cells and passive zones representing conductive components
# such as tabs and busbars, with defined electrical and thermal material properties.
# The model also incorporates natural convection cooling with a heat transfer coefficient
# of 5 W/m²·K and an ambient temperature of 300 K, allowing for realistic thermal
# management analysis.
#
# .. image:: ../../_static/battery_pack_schematic.png
#    :align: center
#    :alt: Battery Pack Schematic

# %%
# Import modules
# --------------
#
# .. note::
#   Importing the following classes offers streamlined access to key solver settings,
#   eliminating the need to manually browse through the full hierarchy of settings APIs structure.

import os

import ansys.fluent.core as pyfluent
from ansys.fluent.core import examples
from ansys.fluent.core.solver import (
    Battery,
    BoundaryConditions,
    CellZoneConditions,
    Contour,
    Controls,
    General,
    Graphics,
    Initialization,
    Materials,
    Mesh,
    ReportDefinitions,
    RunCalculation,
    Solution,
    Vector,
)
from ansys.fluent.visualization import GraphicsWindow, Monitor

# %%
# Launch Fluent in solver mode
# ----------------------------

solver = pyfluent.launch_fluent(
    precision=pyfluent.Precision.DOUBLE,
    mode=pyfluent.FluentMode.SOLVER,
)

# %%
# Read mesh
# ---------

mesh_file = examples.download_file(
    "1P3S_battery_pack.msh",
    "pyfluent/battery_pack",
    save_path=os.getcwd(),
)

solver.settings.file.read_mesh(file_name=mesh_file)

# %%
# Display mesh
# ------------

graphics = Graphics(solver)
mesh = Mesh(solver, new_instance_name="mesh-1")

all_walls = mesh.surfaces_list.allowed_values()

graphics.picture.x_resolution = 650  # Horizontal resolution for clear visualization
graphics.picture.y_resolution = 450  # Vertical resolution matching typical aspect ratio

mesh.surfaces_list = all_walls
mesh.options.edges = True
mesh.display()

graphics.picture.save_picture(file_name="battery_pack_1.png")

# %%
# .. image:: ../../_static/battery_pack_1.png
#    :align: center
#    :alt: Battery Pack Mesh

# %%
# Solver settings
# ---------------

solver_general_settings = General(solver)

solver_general_settings.solver.time = "unsteady-1st-order"

# %%
# Enable Battery Model (NTGK/DCIR)
# --------------------------------
#
# .. note::
#   Using wildcards (`*`, `?`, `|`, etc.) instead of explicit zone lists makes the setup
#   more flexible and scalable. For example:
#
#   - ``cell_*`` → selects all zones starting with ``cell_`` (e.g., ``cell_1``, ``cell_2``, ...)
#   - ``*bar*|*tabzone*`` → selects zones containing either ``bar`` or ``tabzone`` (e.g., ``bar1``, ``n_tabzone_1``,...)

battery_model = Battery(solver)

battery_model.enabled = True
battery_model.echem_model = "ntgk/dcir"
battery_model.eload_condition.eload_settings.eload_type = "specified-system-power"
battery_model.eload_condition.eload_settings.power_value = 200  # W  (Total pack power)

# Conductive zones
battery_model.zone_assignment.active_zone = "cell_*"  # Active cells
battery_model.zone_assignment.passive_zone = "*bar*|*tabzone*"  # Bars zones + tab zones

# External electrical contacts
battery_model.zone_assignment.negative_tab = ["tab_n"]  # Negative terminal
battery_model.zone_assignment.positive_tab = ["tab_p"]  # Positive terminal

# %%
# Define materials
# ----------------

materials = Materials(solver)

# Active material (cells): conductivity via UDS-0 and UDS-1
materials.solid.create("e_material")

materials.solid["e_material"] = {
    "chemical_formula": "e",
    "thermal_conductivity": {"value": 20},  # W/(m·K)
    "uds_diffusivity": {
        "option": "defined-per-uds",
        "uds_diffusivities": {
            "uds-0": {"value": 1000000},  # S/m  (Electronic conductivity)
            "uds-1": {"value": 1000000},  # S/m  (Ionic conductivity)
        },
    },
}

# Passive material (busbars & tabs): high constant conductivity
materials.solid.create("busbar_material")

materials.solid["busbar_material"] = {
    "chemical_formula": "bus",
    "uds_diffusivity": {
        "option": "value",
        "value": 3.541e7,  # S/m  (Copper-like conductivity)
    },
}

# %%
# Assign materials to cell zones
# ------------------------------

cell_zone_conditions = CellZoneConditions(solver)

# Assign e_material to cell_1
cell_zone_conditions.solid["cell_1"] = {"general": {"material": "e_material"}}

# Copy to cell_2 and cell_3
cell_zone_conditions.copy(from_="cell_1", to="cell_*")

# Assign busbar_material to bar1
cell_zone_conditions.solid["bar1"] = {"general": {"material": "busbar_material"}}

# Copy to all passive zones
cell_zone_conditions.copy(from_="bar1", to="*bar*|*tabzone*")
# '*bar*' matches any zone containing 'bar',
# '|*tabzone*' matches any zone containing 'tabzone' → OR logic

# %%
# Boundary conditions
# -------------------

conditions = BoundaryConditions(solver)

# Convection on wall-cell_1
conditions.wall["wall-cell_1"] = {
    "thermal": {
        "thermal_condition": "Convection",
        "heat_transfer_coeff": {"value": 5},  # W/(m²·K)
    }
}

# Copy to all other walls (tabs, busbars, cells)
conditions.copy(from_="wall-cell_1", to="wall*")

# %%
# Define solution controls and monitors
# ---------------------------------------
controls = Controls(solver)

# Disable flow and turbulence equations
controls.equations = {
    "flow": False,
    "kw": False,
}

solution = Solution(solver)

solution.monitor.residual.options.criterion_type = (
    "none"  # No automatic convergence check
)

# %%
# Define report definitions
# -------------------------
definitions = ReportDefinitions(solver)

# Surface report: voltage at positive tab (area-weighted average)
definitions.surface["voltage_surface_areaavg"] = {
    "report_type": "surface-areaavg",
    "field": "passive-zone-potential",
    "surface_names": ["tab_p"],
    "create_report_file": True,
    "create_report_plot": True,
}

# Format plot axes
report_plot = solver.settings.solution.monitor.report_plots[
    "voltage_surface_areaavg-rplot"
]
report_plot.axes.x.number_format.precision = 0  # Integer time steps
report_plot.axes.y.number_format.precision = 2  # 2 decimal places for voltage

# Volume report: maximum temperature in all cell zones
definitions.volume["volume_max_temp"] = {
    "report_type": "volume-max",
    "field": "temperature",
    "cell_zones": "*cell|*bar*|*tabzone*",
    "create_report_file": True,
    "create_report_plot": True,
}

# Format plot axes
report_plot_1 = solver.settings.solution.monitor.report_plots["volume_max_temp-rplot"]
report_plot_1.axes.x.number_format.precision = 0
report_plot_1.axes.y.number_format.precision = 2

# %%
# Initialize solution
# -------------------
initialize = Initialization(solver)
initialize.standard_initialize()

# %%
# Transient controls
# ------------------
Transient_controls = solver.settings.solution.run_calculation.transient_controls

Transient_controls.time_step_count = 50  # Number of time steps
Transient_controls.time_step_size = 30  # s  (30 s per step)

calculation = RunCalculation(solver)
calculation.calculate()  # Run transient simulation

# %%
# Post-processing
# ---------------

# Current density vector plot
vector = Vector(solver, new_instance_name="current-magnitude-vector")

vector.vector_field = "current-density-j"  # A/m²  (Current density vector)
vector.field = "current-magnitude"  # A/m²  (Magnitude for coloring)
vector.surfaces_list = ["tab_n", "tab_p", "wall*"]
vector.options.vector_style = "arrow"
vector.options.scale = 0.03  # Scale factor for visibility
vector.vector_opt.fixed_length = True  # Uniform arrow length
vector.display()

graphics.views.restore_view(view_name="isometric")
graphics.picture.save_picture(file_name="battery_pack_2.png")

# %%
# .. image:: ../../_static/battery_pack_2.png
#    :align: center
#    :alt: Current density vector plot

# Temperature contour
temp_contour = Contour(solver, new_instance_name="temp_contour")

temp_contour.field = "temperature"  # K
temp_contour.surfaces_list = ["tab_n", "tab_p", "wall*"]
temp_contour.colorings.banded = True
temp_contour.display()

graphics.views.restore_view(view_name="isometric")
graphics.picture.save_picture(file_name="battery_pack_3.png")

# %%
# .. image:: ../../_static/battery_pack_3.png
#    :align: center
#    :alt: Temperature contour

# Joule heat source contour
joule_contour = Contour(solver, new_instance_name="joule_heating_contour")

joule_contour.field = "battery-joule-heat-source"  # W/m³
joule_contour.surfaces_list = ["tab_n", "tab_p", "wall*"]
joule_contour.colorings.banded = True
joule_contour.display()

graphics.views.restore_view(view_name="isometric")
graphics.picture.save_picture(file_name="battery_pack_4.png")

# %%
# .. image:: ../../_static/battery_pack_4.png
#    :align: center
#    :alt: Joule heat source contour

# Total heat source contour
total_heat_contour = Contour(solver, new_instance_name="total_heating_contour")

total_heat_contour.field = "total-heat-source"  # W/m³  (Joule + reaction heat)
total_heat_contour.surfaces_list = ["tab_n", "tab_p", "wall*"]
total_heat_contour.colorings.banded = True
total_heat_contour.display()

graphics.views.restore_view(view_name="isometric")
graphics.picture.save_picture(file_name="battery_pack_5.png")

# %%
# .. image:: ../../_static/battery_pack_5.png
#    :align: center
#    :alt: Total heat source contour

# %%
# Display monitor plots
# ---------------------

plot_window = GraphicsWindow(solver)

voltage_rplot = Monitor(solver, monitor_set_name="voltage_surface_areaavg-rplot")
plot_window.add_plot(voltage_rplot, position=(0, 0))

temp_rplot = Monitor(solver, monitor_set_name="volume_max_temp-rplot")
plot_window.add_plot(temp_rplot, position=(1, 1))

plot_window.renderer = "matplotlib"
plot_window.show()

# %%
# .. image:: ../../_static/battery_pack_8.png
#    :align: center
#    :alt: Temperature and voltage Monitor Plot

# %%
# Save case and data
# ------------------------
solver.settings.file.write_case_data(file_name="1P3S_Battery_Pack")

# %%
# Close Fluent
# -------------

solver.exit()

# %%
# Summary
# -------
#
# This example provides a complete and reproducible PyFluent workflow
# for simulating a multi-cell battery pack using the Battery Model with
# the NTGK sub-model. It demonstrates how to set up active and passive
# zones, apply constant power discharge, use UDS-based dual-path
# conductivity, and monitor voltage and temperature behavior during
# operation. It serves as a template for battery pack analysis,
# applicable to designs ranging from small modules to full EV packs.

# %%
# References:
# -----------
# [1] Simulating a 1P3S Battery Pack Using the Battery Model, `Ansys Fluent documentation​ <https://ansyshelp.ansys.com/public/account/secured?returnurl=/Views/Secured/corp/v252/en/flu_tg/flu_bat_tutorial_pack.html>`_.

# sphinx_gallery_thumbnail_path = '_static/battery_pack_2.png'



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Render ui UI in Jupyter notebook."""

from ansys.fluent.core.ui.utils import (
    _parse_path,
    _render_widget_from_props_generic,
    _safe_get_properties,
)

try:
    import ipywidgets as widgets
except ModuleNotFoundError as exc:
    raise ModuleNotFoundError(
        "Missing dependencies, use 'pip install ansys-fluent-core[ui-jupyter]' to install them."
    ) from exc

from ansys.fluent.core.solver.flobject import (
    BaseCommand,
    Group,
    NamedObject,
)


def set_auto_refresh():
    """Refreshes the UI w.r.t. server state for each command execution or parameter invocation."""
    raise NotImplementedError("This is yet to be implemented in jupyter environment.")


def _render_widgets_from_props(settings_obj, label, props):
    """Render widget using pre-fetched props instead of repeated calls."""
    return _render_widget_from_props_generic(settings_obj, label, props, widgets)


def _param_ui(settings_obj, props):
    label = props["python_name"].replace("_", " ").capitalize()

    def get_fn():
        try:
            return getattr(props["parent"], props["python_name"])
        except AttributeError:
            return props["parent"][props["obj_name"]]

    def set_fn(v):
        return setattr(settings_obj.parent, props["python_name"], v)

    widget = _render_widgets_from_props(get_fn(), label, props)
    output = widgets.Output()
    with output:
        output.clear_output()
        print(_parse_path(settings_obj))
    if hasattr(widget, "_is_list_text"):
        typ, parse_csv = widget._is_list_text

        def commit_text_list(change):
            if change["name"] == "value":
                raw = change["new"]
                vals = (
                    [typ(v.strip()) for v in raw.split(",") if v.strip()]
                    if parse_csv
                    else list(change["new"])
                )
                with output:
                    output.clear_output()
                    set_fn(vals)
                    print(f"{_parse_path(settings_obj)} = {vals}")

        widget.observe(commit_text_list)
    else:

        def on_change(change):
            if change["name"] == "value":
                with output:
                    output.clear_output()
                    try:
                        set_fn(change["new"])
                        print(f"{_parse_path(settings_obj)} = {change['new']}")
                    except Exception as e:
                        print(f"Error setting {label}: {e}")

        widget.observe(on_change)
    return widgets.VBox([widget, output])


def _command_ui(func, props):
    """
    Renders input widgets for function arguments based on .argument_names()
    and executes func(**kwargs) on button click.
    """

    # Get argument names from the function object
    if not hasattr(func, "argument_names"):
        return widgets.HTML("Command has no 'argument_names()'.")

    arg_names = func.argument_names
    arg_widgets = {}
    controls = []
    for name in arg_names:
        child_obj = getattr(func, name)
        child_props = _safe_get_properties(child_obj)
        widget = _render_widgets_from_props(child_obj, name, child_props)
        arg_widgets[name] = widget
        controls.append(widget)

    # Run button
    button = widgets.Button(
        description=f"Run {props['python_name']}", button_style="success"
    )
    output = widgets.Output()
    with output:
        output.clear_output()
        print(_parse_path(func))

    def on_click(_):
        kwargs = {name: w.value for name, w in arg_widgets.items()}
        with output:
            output.clear_output()
            try:
                func(**kwargs)
                kwargs_str = "("
                for k, v in kwargs.items():
                    if type(v) is str:
                        if v != "":
                            kwargs_str += f"{k}='{v}', "
                    else:
                        kwargs_str += f"{k}={v}, "
                print(f"{_parse_path(func)}" + kwargs_str.strip()[:-1] + ")")
            except Exception as e:
                print("Error:", e)

    button.on_click(on_click)
    return widgets.VBox(controls + [button, output])


def settings_ui(obj, indent=0):
    """Render settings objects into ui graphics."""
    props = _safe_get_properties(obj)
    if isinstance(obj, (Group, NamedObject)):
        if isinstance(obj, Group):
            command_names = obj.get_active_command_names()
            child_names = obj.get_active_child_names() + command_names
        else:
            command_names = obj.command_names
            child_names = list(obj) + command_names
        accordions = []
        for child_name in child_names:

            def lazy_loader(name=child_name, parent=obj, lvl=indent + 1):
                try:
                    child_obj = getattr(parent, name)
                except AttributeError:
                    child_obj = parent[name]
                return settings_ui(child_obj, lvl)

            acc = widgets.Accordion(children=[widgets.HTML("Loading...")])
            if child_name in command_names:
                acc.set_title(0, f"⚡ {child_name}")
            else:
                acc.set_title(0, child_name)

            def on_selected(change, loader=lazy_loader, accordion=acc):
                if change["name"] == "selected_index" and change["new"] == 0:
                    if isinstance(accordion.children[0], widgets.HTML):
                        accordion.children = [loader()]

            acc.observe(on_selected, names="selected_index")
            accordions.append(acc)

        return widgets.VBox(accordions)

    else:
        if isinstance(obj, BaseCommand):
            return (
                widgets.VBox([_command_ui(obj, props)])
                if props["is_active"]
                else widgets.HTML("")
            )
        else:
            return (
                widgets.VBox([_param_ui(obj, props)])
                if props["is_active"]
                else widgets.HTML("")
            )



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

""".. _single_battery_cell_simulation:

Single Battery Cell Using MSMD Battery Model Simulation
----------------------------------------------------------------------------
"""

#######################################################################################
# Problem Description:
# =====================================================================================
# Simulate a 14.6 Ah lithium-ion battery with a LiMn₂O₄ cathode and graphite anode using
# the Multi-Scale Multi-Domain (MSMD) battery model in Ansys Fluent. Evaluate the battery's
# electrochemical and thermal performance under various discharge rates (e.g., 0.5C, 1C, 5C)
# and operating conditions, including normal operation, pulse discharge, and short-circuit
# scenarios. Key outputs include voltage, temperature, and state of charge.
#
# .. image:: ../../_static/Single_Battery_cell_model.png
#    :align: center
#    :alt: Single Battery Cell Model

#######################################################################################
# Import modules
# =====================================================================================
import os

import ansys.fluent.core as pyfluent
from ansys.fluent.core import FluentMode, Precision, examples

#######################################################################################
# Launch Fluent session
# =====================================================================================
# Launch a Fluent solver session with required parameters
solver = pyfluent.launch_fluent(
    precision=Precision.DOUBLE, processor_count=4, mode=FluentMode.SOLVER
)

#######################################################################################
# Download the mesh file
# =====================================================================================
# Download the battery mesh file and save it to the current working directory.
unit_battery_mesh = examples.download_file(
    "unit_battery.msh.h5",
    "pyfluent/battery_thermal_simulation",
    save_path=os.getcwd(),
)

#######################################################################################
# Read and display mesh
# =====================================================================================
#
# .. note::
#   Graphics commands like restore_view and save_picture require GUI mode.
solver.settings.file.read_case(file_name=unit_battery_mesh)

# Get all the available wall boundary surfaces
all_walls = solver.settings.setup.boundary_conditions.wall.get_object_names()
mesh_object = solver.settings.results.graphics.mesh.create("mesh-1")
mesh_object.surfaces_list = all_walls
mesh_object.options.edges = True
mesh_object.display()

graphics_object = solver.settings.results.graphics
graphics_object.picture.x_resolution = 650
graphics_object.picture.y_resolution = 450
graphics_object.views.restore_view(view_name="isometric")
graphics_object.picture.save_picture(file_name="Single_Battery_Cell_Mesh.png")

# %%
# .. image:: ../../_static/Single_Battery_Cell_Mesh.png
#    :align: center
#    :alt: Single Battery Cell Mesh

#######################################################################################
# Configure solver settings for battery model
# =====================================================================================
# Use an unsteady first-order time solver for transient behavior.
solver.settings.setup.general.solver.time = "unsteady-1st-order"

#######################################################################################
# Enable the battery model
# =====================================================================================
# Activate the NTGK/DCIR model with a nominal cell capacity of 14.6 Ah.
# Enable Joule heat in passive zones and define zones and terminals.
# For a detailed guide on setting up a single battery cell,refer to the Reference_ [3].

battery = solver.settings.setup.models.battery
battery.enabled = True
battery.echem_model = "ntgk/dcir"
battery.zone_assignment.active_zone = ["e_zone"]
battery.zone_assignment.passive_zone = ["tab_nzone", "tab_pzone"]
battery.zone_assignment.negative_tab = ["tab_n"]
battery.zone_assignment.positive_tab = ["tab_p"]

#######################################################################################
# Define materials for cell and tabs
# =====================================================================================
#
# .. note::
#   Chemical formula values are arbitrary identifiers for demonstration.
#
# Material definition for battery cell, positive tab and negative tab. User define
# scalars are defined for e-material and positive material to specify the
# electric conductivity with ``defined-per-uds`` and ``constant`` option respectively.
#
materials = [
    {
        "name": "e_material",
        "chemical_formula": "e",
        "density": 2092,
        "specific_heat": 678,
        "thermal_conductivity": 18.2,
        "uds_diffusivity": {
            "option": "defined-per-uds",
            "uds-0": 1190000,
            "uds-1": 983000,
        },
    },
    {
        "name": "p_material",
        "chemical_formula": "pmat",
        "density": 8978,
        "specific_heat": 381,
        "thermal_conductivity": 387.6,
        "uds_diffusivity": {"option": "constant", "value": 10000000},
    },
    {
        "name": "n_material",
        "chemical_formula": "nmat",
        "density": 8978,
        "specific_heat": 381,
        "thermal_conductivity": 387.6,
    },
]

solids = solver.settings.setup.materials.solid
for mat in materials:
    solids.create(mat["name"])
    solids[mat["name"]].chemical_formula = mat["chemical_formula"]
    solids[mat["name"]].density.value = mat["density"]
    solids[mat["name"]].specific_heat.value = mat["specific_heat"]
    solids[mat["name"]].thermal_conductivity.value = mat["thermal_conductivity"]
    if "uds_diffusivity" in mat:
        solids[mat["name"]].uds_diffusivity = {
            "option": mat["uds_diffusivity"]["option"]
        }
        if mat["uds_diffusivity"]["option"] == "defined-per-uds":
            solids[mat["name"]].uds_diffusivity.uds_diffusivities["uds-0"].value = mat[
                "uds_diffusivity"
            ]["uds-0"]
            solids[mat["name"]].uds_diffusivity.uds_diffusivities["uds-1"].value = mat[
                "uds_diffusivity"
            ]["uds-1"]
        else:
            solids[mat["name"]].uds_diffusivity.value = mat["uds_diffusivity"]["value"]

#######################################################################################
# Assign materials to cell zones
# =====================================================================================
# Map materials to respective zones.

cell_zones = [
    ("e_zone", "e_material"),
    ("tab_nzone", "n_material"),
    ("tab_pzone", "p_material"),
]

for zone, material in cell_zones:
    solver.settings.setup.cell_zone_conditions.solid[zone].general.material = material

#######################################################################################
# Define boundary conditions
# =====================================================================================
# Set convective heat transfer on external surfaces.

wall = solver.settings.setup.boundary_conditions.wall
wall["wall_active"].thermal.thermal_condition = "Convection"
wall["wall_active"].thermal.heat_transfer_coeff.value = 5

# API to copy similar boundary condition
solver.settings.setup.boundary_conditions.copy(
    from_="wall_active", to=["wall_n", "wall_p"]
)

#######################################################################################
# Configure solution settings
# =====================================================================================
# Disable flow and turbulence equations, since residual criteria are set to ``none``

solver.settings.solution.controls.equations["flow"] = False
solver.settings.solution.controls.equations["kw"] = False

solver.settings.solution.monitor.residual.options.criterion_type = "none"

#######################################################################################
# Create report definitions
# =====================================================================================
# Monitor average voltage and maximum temperature.

avg_surface_voltage_report_def = (
    solver.settings.solution.report_definitions.surface.create("surface_voltage")
)
avg_surface_voltage_report_def.report_type = "surface-areaavg"
avg_surface_voltage_report_def.field = "passive-zone-potential"
avg_surface_voltage_report_def.surface_names = ["tab_p"]

max_temp_report_def = solver.settings.solution.report_definitions.volume.create(
    "max_temperature"
)
max_temp_report_def.report_type = "volume-max"
max_temp_report_def.field = "temperature"
max_temp_report_def.cell_zones = ["e_zone", "tab_nzone", "tab_pzone"]

surf_voltage_report_files = solver.settings.solution.monitor.report_files.create(
    "surface_voltage_file"
)
surf_voltage_report_files.report_defs = ["flow-time", "surface_voltage"]
surf_voltage_report_files.file_name = "ntgk-1c.out"
surf_voltage_report_files.print = True

max_temp_report_file = solver.settings.solution.monitor.report_files.create(
    "max_temperature_file"
)
max_temp_report_file.report_defs = ["max_temperature"]
max_temp_report_file.file_name = "max-temp-1c.out"
max_temp_report_file.print = True

report_plots = solver.settings.solution.monitor.report_plots

voltage_plot = report_plots.create("surface_voltage_plot")
voltage_plot.report_defs = ["surface_voltage"]
voltage_plot.print = True
voltage_plot.axes.x.number_format.precision = 0
voltage_plot.axes.y.number_format.precision = 2

temp_plot = report_plots.create("max_temperature_plot")
temp_plot.report_defs = ["max_temperature"]
temp_plot.print = True
temp_plot.axes.x.number_format.precision = 0
temp_plot.axes.y.number_format.precision = 2

#######################################################################################
# Run the simulation
# =====================================================================================

solver.settings.solution.initialization.standard_initialize()
transient_controls = solver.settings.solution.run_calculation.transient_controls
transient_controls.time_step_size = 30
transient_controls.time_step_count = 100
solver.settings.solution.run_calculation.calculate()

#######################################################################################
# Post-process results
# =====================================================================================
# Generate contour and vector plots.

contours = solver.settings.results.graphics.contour
contour_list = [
    {
        "name": "contour-phi+",
        "field": "cathode-potential",
        "surfaces": ["wall_active"],
        "file_name": "Single_Battery_Cell_1.png",
    },
    {
        "name": "contour-phi-",
        "field": "anode-potential",
        "surfaces": ["wall_active"],
        "file_name": "Single_Battery_Cell_2.png",
    },
    {
        "name": "contour-phi-passive",
        "field": "passive-zone-potential",
        "surfaces": ["tab_n", "tab_p", "wall_n", "wall_p"],
        "file_name": "Single_Battery_Cell_3.png",
    },
    {
        "name": "contour-temp",
        "field": "temperature",
        "surfaces": ["wall_p", "wall_active", "tab_p", "tab_n", "wall_n"],
        "file_name": "Single_Battery_Cell_4.png",
    },
]

# Create, display, and save contour plots
for contour in contour_list:
    # Create the contour
    contours.create(contour["name"])
    current = contours[contour["name"]]
    current.field = contour["field"]
    current.surfaces_list = contour["surfaces"]
    current.range_options.compute()

    # Set the  view
    graphics_object.views.restore_view(view_name="front")
    #  display the  current contour
    current.display()
    # Save the contour plot as an image
    graphics_object.picture.save_picture(file_name=contour["file_name"])

# Create and configure vector plot
vector_plot = solver.settings.results.graphics.vector.create("vector-current_density")
vector_plot.vector_field = "current-density-j"
vector_plot.field = "current-magnitude"
vector_plot.surfaces_list = ["wall_n", "wall_p", "wall_active", "tab_n", "tab_p"]
vector_plot.options.vector_style = "arrow"
vector_plot.range_options.compute()
# Set view, display, and save the vector plot image
graphics_object.views.restore_view(view_name="front")
vector_plot.display()
graphics_object.picture.save_picture(file_name="Single_Battery_Cell_5.png")
# Save case file for ROM simulation
solver.settings.file.write_case(file_name="unit_battery.cas.h5")
# Save case and data for short circuit simulation
solver.settings.file.write_case_data(file_name="ntgk")  # Save case data

# %%
# .. image:: ../../_static/Single_Battery_Cell_1.png
#    :align: center
#    :alt:  Anode Potential Contour(1C)
#
# .. image:: ../../_static/Single_Battery_Cell_2.png
#    :align: center
#    :alt:  Cathode Potential Contour(1C)
#
#
# .. image:: ../../_static/Single_Battery_Cell_3.png
#    :align: center
#    :alt:  Passive Zone Potential Contour(1C)
#
#
# .. image:: ../../_static/Single_Battery_Cell_4.png
#    :align: center
#    :alt:  Static Temperature Contour(1C)
#
#
# .. image:: ../../_static/Single_Battery_Cell_5.png
#    :align: center
#    :alt:  Current Density Vector(1C)
#


#######################################################################################
# Run simulations at different C-rates
# =====================================================================================
# Simulate at 0.5C and 5C discharge rates with adjusted time steps.

solver.settings.setup.models.battery.eload_condition.eload_settings.crate_value = 0.5

# Get report files
report_files = solver.settings.solution.monitor.report_files

# Update report file names for 0.5 c rate simulation for existing report files
report_files["surface_voltage_file"].file_name = "ntgk-0.5c.out"
report_files["max_temperature_file"].file_name = "max-temp-0.5c.out"
solver.settings.solution.initialization.standard_initialize()
solver.settings.solution.run_calculation.transient_controls.time_step_count = 230
solver.settings.solution.run_calculation.calculate()

solver.settings.setup.models.battery.eload_condition.eload_settings.crate_value = 5

# Update report file names for 5 c rate simulation for existing report files
report_files["surface_voltage_file"].file_name = "ntgk-5c.out"
report_files["max_temperature_file"].file_name = "max-temp-5c.out"
solver.settings.solution.initialization.standard_initialize()
solver.settings.solution.run_calculation.transient_controls.time_step_count = 23
solver.settings.solution.run_calculation.calculate()

#######################################################################################
# Reduced Order Method (ROM) setup
# =====================================================================================
# Apply ROM for computational efficiency.

solver.settings.file.read_case(file_name="unit_battery.cas.h5")
solver.settings.solution.initialization.standard_initialize()
solver.settings.solution.run_calculation.transient_controls.time_step_size = 30
solver.settings.solution.run_calculation.transient_controls.time_step_count = 3
solver.settings.solution.run_calculation.calculate()

solver.settings.setup.models.battery.solution_method = "msmd-rom"
solver.settings.setup.models.battery.solution_option.option_settings.number_substeps = (
    10
)
solver.settings.solution.run_calculation.transient_controls.time_step_size = 30
solver.settings.solution.run_calculation.transient_controls.time_step_count = 100
solver.settings.solution.run_calculation.calculate()

# Generate contour and vector plots for ROM results.
contours = solver.settings.results.graphics.contour
contour_list = [
    {
        "name": "contour_cathode_potential",
        "field": "cathode-potential",
        "surfaces": ["wall_active"],
    },
    {
        "name": "contour_anode_potential",
        "field": "anode-potential",
        "surfaces": ["wall_active"],
    },
    {
        "name": "contour_passive_potential",
        "field": "passive-zone-potential",
        "surfaces": ["tab_n", "tab_p", "wall_n", "wall_p"],
    },
    {
        "name": "contour_temperature",
        "field": "temperature",
        "surfaces": ["wall_p", "wall_active", "tab_p", "tab_n", "wall_n"],
    },
]

for contour in contour_list:
    contours.create(contour["name"])
    contours[contour["name"]].field = contour["field"]
    contours[contour["name"]].surfaces_list = contour["surfaces"]
    contours[contour["name"]].range_options.compute()

vectors = solver.settings.results.graphics.vector.create("vector-current_density")
vectors.vector_field = "current-density-j"
vectors.field = "current-magnitude"
vectors.surfaces_list = [
    "wall_n",
    "wall_p",
    "wall_active",
    "tab_n",
    "tab_p",
]
vectors.options.vector_style = "arrow"
vectors.range_options.compute()

# Set view, display, and save the vector plot image
graphics_object.views.restore_view(view_name="front")
vectors.display()
graphics_object.picture.save_picture(file_name="Single_Battery_Cell_6.png")

# %%
# .. image:: ../../_static/Single_Battery_Cell_6.png
#    :align: center
#    :alt: Current Magnitude (1C)
# Vector current density for ROM model (faster with identical results).

#######################################################################################
# Simulate short-circuit
# =====================================================================================
# Apply low external resistance and define a short-circuit region.

solver.settings.file.read_case(file_name="ntgk.cas.h5")

solver.settings.setup.models.battery.eload_condition.eload_settings.eload_type = (
    "specified-resistance"
)
solver.settings.setup.models.battery.eload_condition.eload_settings.external_resistance = (
    0.5
)

# Create a new cell register named "register_patch"
patch = solver.settings.solution.cell_registers.create(name="register_patch")
patch.type.option = "hexahedron"

# Configure the hexahedron box
patch.type.hexahedron.inside = True
patch.type.hexahedron.min_point = [-0.01, -0.01, -1.0]
patch.type.hexahedron.max_point = [0.01, 0.02, 1.0]

solver.settings.solution.initialization.standard_initialize()

# Patch initialization
solver.settings.solution.initialization.patch.calculate_patch(
    domain="",
    cell_zones=[],
    registers=["register_patch"],
    variable="battery-short-resistance",
    reference_frame="Relative to Cell Zone",
    use_custom_field_function=False,
    custom_field_function_name="",
    value=5e-07,
)

solver.settings.solution.run_calculation.transient_controls.time_step_size = 1
solver.settings.solution.run_calculation.transient_controls.time_step_count = 5
solver.settings.solution.run_calculation.calculate()

solver.settings.file.write_case_data(file_name="ntgk_short_circuit.cas.h5")

solver.settings.results.report.surface_integrals.area_weighted_avg(
    report_of="passive-zone-potential", surface_names=["tab_p"], write_to_file=False
)
solver.settings.results.report.volume_integrals.volume_integral(
    cell_function="total-current-source", cell_zones=["e_zone"], write_to_file=False
)

vector = solver.settings.results.graphics.vector
vector_negative = vector.create("vector_negative_current")
vector_negative.vector_field = "current-density-jn"
vector_negative.field = "current-magnitude"
vector_negative.surfaces_list = [
    "wall_n",
    "wall_p",
    "wall_active",
]
vector_negative.options.vector_style = "arrow"
vector_negative.range_options.compute()
graphics_object.views.restore_view(view_name="front")
vector_negative.display()
graphics_object.picture.save_picture(file_name="Single_Battery_Cell_9.png")


# %%
# .. image:: ../../_static/Single_Battery_Cell_9.png
#    :align: center
#    :alt: Negative Current Vector Plot
# Negative current vector plot after short circuit.

vector_positive = vector.create("vector_positive_current")
vector_positive.vector_field = "current-density-jp"
vector_positive.field = "current-magnitude"
vector_positive.surfaces_list = [
    "wall_n",
    "wall_p",
    "wall_active",
]
vector_positive.options.vector_style = "arrow"
vector_positive.range_options.compute()

graphics_object.views.restore_view(view_name="front")
vector_positive.display()
graphics_object.picture.save_picture(file_name="Single_Battery_Cell_10.png")

# %%
# .. image:: ../../_static/Single_Battery_Cell_10.png
#    :align: center
#    :alt: Positive Current Vector Plot
# Positive current vector plot after short circuit.

temp_contour = solver.settings.results.graphics.contour.create("temperature-contour")
temp_contour.field = "temperature"
temp_contour.surfaces_list = all_walls
temp_contour.range_options.compute()

graphics_object.views.restore_view(view_name="front")
temp_contour.display()
graphics_object.picture.save_picture(file_name="Single_Battery_Cell_11.png")

# %%
# .. image:: ../../_static/Single_Battery_Cell_11.png
#    :align: center
#    :alt: Temperature Contour
# Temperature contour plot after short circuit.

#######################################################################################
# Close the solver
# =====================================================================================
solver.exit()

#######################################################################################
# References:
# =====================================================================================
# [1] U. S. Kim et al, "Effect of electrode configuration on the thermal behavior of
# a lithium-polymer battery", Journal of Power Sources, Volume 180 (2), pages 909-916, 2008.
#
# [2] U. S. Kim, et al., "Modeling the Dependence of the Discharge Behavior of a Lithium-Ion Battery
# on the Environmental Temperature", J. of Electrochemical Soc., Volume 158 (5), pages A611-A618, 2011.
#
# .. _Reference:
# [3] Simulating a Single Battery Cell Using the MSMD Battery Model, `Ansys Fluent documentation​ <https://ansyshelp.ansys.com/public/account/secured?returnurl=/Views/Secured/corp/v252/en/flu_tg/flu_bat_tutorial_cell.html>`_.

# sphinx_gallery_thumbnail_path = '_static/Single_Battery_Cell_4.png'



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

""".. _Solidification_Simulation:

Modeling Solidification
-----------------------
"""

# %%
# Objective
# ---------
#
# This example models solidification in a Czochralski
# crystal growth process using PyFluent. The simulation involves setting
# up a two-dimensional axisymmetric swirl configuration with gravity and
# activating the Solidification and Melting model to capture phase change
# behavior. Pull velocities are applied to represent continuous casting,
# while Marangoni convection is incorporated through a surface tension
# gradient at the melt surface. The workflow begins with a steady state
# conduction solution to establish a realistic initial thermal field.
# Subsequently, transient flow and heat transfer are enabled to capture
# the combined effects of natural and Marangoni convection. To achieve
# this, custom field functions and patching techniques are employed for
# efficient definition of pull velocities and initialization of the
# simulation.

# %%
# Problem Description
# -------------------
#
# The problem involves simulating a two-dimensional axisymmetric bowl
# containing liquid metal. The liquid enters from the bottom at a
# temperature of 1300 K, while solidified material is continuously
# pulled from the top at a rate of 0.001 m/s. The crystal undergoes
# rotation at an angular velocity of 1 rad/s, which induces a
# swirling flow within the melt. Heat loss from the upper region
# of the crystal promotes solidification of the liquid metal.
# Additionally, Marangoni convection develops at the free surface
# due to temperature-dependent variations in surface tension.
# The solidification process is modeled with zero mushy zone width,
# where both the solidus and liquidus temperatures are 1150 K.
# The flow incorporates an axial pull velocity of 0.001 m/s and
# a swirl velocity defined by ω × r, with ω = 1 rad/s.
# The surface tension gradient responsible for Marangoni stress
# is given by dσ/dT = –0.00036 N/m·K, and gravity acts in the
# axial (X) direction with a magnitude of –9.81 m/s².
#
# .. image:: ../../_static/modeling_solidification.png
#    :align: center
#    :alt: Solidification in Czochralski Model

# %%
# Import modules
# --------------
#
# .. note::
#   Importing the following classes offer streamlined access to key solver settings,
#   eliminating the need to manually browse through the full settings structure.

import os

import ansys.fluent.core as pyfluent
from ansys.fluent.core import examples
from ansys.fluent.core.solver import (
    BoundaryConditions,
    Contour,
    Controls,
    General,
    Graphics,
    Initialization,
    Materials,
    Mesh,
    Methods,
    Results,
    RunCalculation,
    Setup,
    Solution,
    VelocityInlet,
)

# %%
# Launch Fluent session in solver mode
# ------------------------------------
solver = pyfluent.launch_fluent(
    precision=pyfluent.Precision.DOUBLE,
    mode="solver",
    dimension=pyfluent.Dimension.TWO,
)

# %%
# Download mesh file
# ------------------
mesh_file = examples.download_file(
    "solid.msh",
    "pyfluent/solidification",
    save_path=os.getcwd(),
)

solver.settings.file.read_mesh(file_name=mesh_file)

# %%
# Display mesh
# ------------
graphics = Graphics(solver)
mesh = Mesh(solver, new_instance_name="mesh-1")
boundary_conditions = BoundaryConditions(solver)

graphics.picture.x_resolution = 650  # Horizontal resolution for clear visualization
graphics.picture.y_resolution = 450  # Vertical resolution matching typical aspect ratio

all_walls = mesh.surfaces_list.allowed_values()

mesh.surfaces_list = all_walls
mesh.options.edges = True
mesh.display()

graphics.picture.save_picture(file_name="modeling_solidification_1.png")

# %%
# .. image:: ../../_static/modeling_solidification_1.png
#    :align: center
#    :alt: Mesh

# %%
# Configure solver
# ----------------
solver_general_settings = General(solver)

solver_general_settings.solver.two_dim_space = "swirl"

solver_general_settings.operating_conditions.gravity = {
    "enable": True,
    "components": [-9.81],
}

# %%
# Enable models
# -------------
#
# .. note::
#   Energy is auto-enabled with solidification

setup = Setup(solver)

setup.models.viscous.model = "laminar"

# Enable the Solidification/Melting
solver.tui.define.models.solidification_melting(
    "yes", "constant", "100000", "yes", "no"  # 100000 for the Mushy Zone Constant.
)

# %%
# Define material
# ---------------
materials = Materials(solver)

materials.database.copy_by_name(type="fluid", name="air", new_name="liquid-metal")

materials.fluid["liquid-metal"] = {
    "density": {
        "polynomial": {
            "coefficients": [
                8000,
                -0.1,
            ],  # [Density (kg/m³), Linear temp coefficient (kg/(m³·K))]
            "function_of": "temperature",
        },
        "option": "polynomial",
    },
    "viscosity": {"value": 0.00553},  # Pa·s
    "specific_heat": {"value": 680.0},  # J/kg·K
    "thermal_conductivity": {"value": 30.0},  # W/m·K
    "melting_heat": {"value": 100000.0},  # J/kg
    "tsolidus": {"value": 1150.0},  # K
    "tliqidus": {"value": 1150.0},  # K
}

# Assign material to fluid zone
setup.cell_zone_conditions.fluid["fluid"] = {"general": {"material": "liquid-metal"}}

# %%
# Boundary conditions
# -------------------

# Inlet: liquid injection
inlet = VelocityInlet(solver, name="inlet")

inlet.momentum.velocity_magnitude.value = 0.00101  # m/s
inlet.thermal.temperature.value = 1300  # K

# Outlet: solid pull-out (velocity inlet with axial + swirl)
outlet = VelocityInlet(solver, name="outlet")

outlet.momentum.velocity_specification_method = "Components"
outlet.momentum.swirl_angular_velocity = 1  # rad/s
outlet.momentum.velocity_components = [0.001, 0, 0]  # Axial = 0.001 m/s
outlet.thermal.temperature.value = 500  # K


conditions = BoundaryConditions(solver)

# Bottom wall: fixed temperature
conditions.wall["bottom-wall"] = {
    "thermal": {"thermal_condition": "Temperature", "temperature": 1300}  # K
}

# Free surface: Marangoni stress + convection
conditions.wall["free-surface"] = {
    "momentum": {
        "shear_condition": "Marangoni Stress",
        "surface_tension_gradient": -0.00036,  # N/m·K
    },
    "thermal": {
        "thermal_condition": "Convection",
        "free_stream_temp": 1500,  # K
        "heat_transfer_coeff": 100,  # W/m²·K
    },
}

# Side wall: fixed temperature
conditions.wall["side-wall"] = {
    "thermal": {"thermal_condition": "Temperature", "temperature": 1400}  # K
}

# Solid wall: rotating + cold
conditions.wall["solid-wall"] = {
    "momentum": {
        "wall_motion": "Moving Wall",
        "velocity_spec": "Rotational",
        "rotation_speed": 1,  # rad/s
    },
    "thermal": {"thermal_condition": "Temperature", "temperature": 500},  # K
}

# %%
# Solution methods
# ----------------
methods = Methods(solver)

methods.p_v_coupling.flow_scheme = "Coupled"
methods.spatial_discretization.discretization_scheme = {"pressure": "presto!"}
methods.pseudo_time_method.formulation = {"coupled_solver": "global-time-step"}

# %%
# Disable flow equations
# ----------------------
controls = Controls(solver)

controls.equations = {"flow": False, "w-swirl": False}


# %%
# Initialize flow field
# ---------------------
initialize = Initialization(solver)
initialize.hybrid_initialize()

# %%
# Define custom field function
# ----------------------------
results = Results(solver)

results.custom_field_functions.create(
    name="omegar", custom_field_function="1 * radial_coordinate"  # ω = 1 rad/s
)

# %%
# Patch pull velocities
# ---------------------

# Axial pull velocity = 0.001 m/s
initialize.patch.calculate_patch(
    cell_zones=["fluid"],
    variable="x-pull-velocity",
    reference_frame="Relative to Cell Zone",
    use_custom_field_function=True,
    custom_field_function_name="omegar",
    value=0.001,
)

# Swirl pull velocity = ω × r
initialize.patch.calculate_patch(
    cell_zones=["fluid"],
    variable="z-pull-velocity",
    reference_frame="Relative to Cell Zone",
    use_custom_field_function=True,
    custom_field_function_name="omegar",
)

# %%
# Pseudo-transient settings
# -------------------------
solver.settings.solution.run_calculation.pseudo_time_settings.time_step_method = {
    "time_step_method": "user-specified",
    "auto_time_size_calc_solid_zone": False,
}


calculation = RunCalculation(solver)
calculation.iterate(iter_count=20)

# %%
# Post-processing
# ---------------
temp_contour = Contour(solver, new_instance_name="temperature_contour")

temp_contour.coloring.option = "banded"
temp_contour.field = "temperature"
temp_contour.display()

graphics.views.restore_view(view_name="front")
graphics.picture.save_picture(file_name="modeling_solidification_2.png")

# %%
# .. image:: ../../_static/modeling_solidification_2.png
#    :align: center
#    :alt: Temperature Contours


mushy_temp = Contour(solver, new_instance_name="temperature-mushy")
mushy_temp.coloring.option = "banded"
mushy_temp.field = "temperature"
mushy_temp.range_option = {
    "option": "auto-range-off",
    "auto_range_off": {"clip_to_range": True, "minimum": 1100, "maximum": 1200},
}
mushy_temp.display()

graphics.views.restore_view(view_name="front")
graphics.picture.save_picture(file_name="modeling_solidification_3.png")

# %%
# .. image:: ../../_static/modeling_solidification_3.png
#    :align: center
#    :alt: Mushy Zone


# Save steady state case
solver.settings.file.write_case_data(file_name="steady_state")


# %%
# Enable transient flow and heat transfer
# ---------------------------------------
solver_general_settings.solver.time = "unsteady-1st-order"  # First-order implicit

controls.equations = {"flow": True, "w-swirl": True}

controls.under_relaxation = {"delh": 0.1}

# %%
# Transient controls
# ------------------
solutions = Solution(solver)

solutions.run_calculation.transient_controls.time_step_count = 2
solutions.run_calculation.transient_controls.time_step_size = 0.1  # s
solutions.run_calculation.calculate()

# Liquid fraction at t = 0.2 s
liquid_fraction_contour = Contour(solver, new_instance_name="liquid-fraction")

liquid_fraction_contour.field = "liquid-fraction"
liquid_fraction_contour.display()

graphics.picture.save_picture(file_name="modeling_solidification_4.png")

# %%
# .. image:: ../../_static/modeling_solidification_4.png
#    :align: center
#    :alt: Liquid Fraction at t = 0.2 s

# Continue to t = 5.0 s (48 more steps)
solutions.run_calculation.transient_controls.time_step_count = 48

solutions.run_calculation.calculate()

# Liquid fraction at t = 5.0 s
liquid_fraction_contour_t_5_sec = Contour(solver, new_instance_name="liquid-fraction")

liquid_fraction_contour_t_5_sec.field = "liquid-fraction"
liquid_fraction_contour_t_5_sec.display()

graphics.picture.save_picture(file_name="modeling_solidification_5.png")

# %%
# .. image:: ../../_static/modeling_solidification_5.png
#    :align: center
#    :alt: Liquid Fraction at t = 5 s

# Save transient case
solver.settings.file.write_case_data(file_name="unsteady_state")

# %%
# Close session
# -------------
solver.exit()

# %%
# Summary
# -------
#
# In this example, we used PyFluent to model solidification
# in a Czochralski crystal growth process. The workflow begins
# with a steady conduction solution to establish the initial
# thermal field, followed by transient flow to capture natural
# and Marangoni-driven convection. Pull velocities are applied
# through a custom field function (ω × r), and surface tension
# gradients define Marangoni effects. This approach is efficient,
# fully scriptable, and easily adaptable to other casting or
# crystal growth simulations.


# %%
# References:
# -----------
# [1] Modeling Solidification, `Ansys Fluent documentation​ <https://ansyshelp.ansys.com/public/account/secured?returnurl=/Views/Secured/corp/v252/en/flu_tg/x1-37700023.html>`_.

# sphinx_gallery_thumbnail_path = '_static/modeling_solidification_2.png'



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

""".. _catalytic_converter_workflow:

Modeling Flow Through Porous Media - Catalytic Converter
========================================================
"""

# %%
# Introduction
# ------------
#
# Many industrial applications such as filters, catalyst beds, and packing, involve modeling the flow
# through porous media. This tutorial demonstrates the simulation workflow for a catalytic converter using PyFluent.
# The example showcases a complete workflow from geometry import through meshing to solution setup and
# post-processing.
#
# This tutorial demonstrates how to do the following:
#
# * Import CAD geometry for a catalytic converter
# * Set up a watertight meshing workflow with local sizing controls
# * Generate a surface mesh
# * Cap inlets and outlets
# * Extract a fluid region
# * Generate a volume mesh
# * Configure porous media zones for the substrates with appropriate resistances.
# * Set up boundary conditions
# * Set up monitors, report definitions to track mass flow rate.
# * Initialize and run the simulation
# * Postprocessing results
#
# Problem Description
# -------------------
#
# The system consists of a catalytic converter with the following components:
#
# * **Fluid regions**: Primary flow path and substrate regions
# * **Solid regions**: Catalytic substrate material with porous media properties
# * **Sensing elements**: Temperature and flow monitoring sensors
# * **Boundary conditions**: High-temperature inlet flow at 800K and 125 m/s
#
# The simulation models heat transfer through the catalytic substrate using porous
# media with specified inertial and viscous resistance values.
#
# Background
# ----------
#
# Catalytic converters operate by passing exhaust gases through a porous substrate
# coated with catalytic materials. The porous media approach allows modeling of:
#
# * Pressure drop through substrate
# * Heat transfer between gas and solid phases
# * Flow distribution effects in the converter
#
# The substrate is modeled as a porous zone with:
#
# * **Porosity**: 1.0 (fully open channels)
# * **Inertial resistance**: [1000, 1000, 1000] 1/m for pressure drop
# * **Viscous resistance**: [1×10⁶, 1×10⁶, 1000] 1/m² for flow resistance
#
# Setup and Solution
# ------------------
#
# Preparation
# ^^^^^^^^^^^
# Launch Fluent meshing session and initialize the workflow.

# sphinx_gallery_capture_repr = ('_repr_html_', '__repr__')
# sphinx_gallery_thumbnail_path = '_static/catalytic_converter/catalytic_converter_cad_geo.png'

import os
import platform

import ansys.fluent.core as pyfluent
from ansys.fluent.core import (
    Dimension,
    FluentMode,
    FluentVersion,
    Precision,
    UIMode,
    examples,
)
from ansys.fluent.core.solver import (  # noqa: E402
    CellZoneCondition,
    General,
    Graphics,
    Initialization,
    Materials,
    Mesh,
    Models,
    Monitor,
    PressureOutlet,
    ReportDefinitions,
    Results,
    RunCalculation,
    Scene,
    VelocityInlet,
)

# Launch meshing session
meshing = pyfluent.launch_fluent(
    product_version=FluentVersion.v252,
    mode=FluentMode.MESHING,
    ui_mode=UIMode.GUI,
    processor_count=4,
    precision=Precision.DOUBLE,
    dimension=Dimension.THREE,
)

# %%
# Meshing Workflow
# ^^^^^^^^^^^^^^^^
# Set up the watertight geometry workflow for complex multi-region geometry.

# Initialize watertight geometry workflow

workflow = meshing.workflow
workflow.InitializeWorkflow(WorkflowType=r"Watertight Geometry")

# %%
# Import Geometry
# ~~~~~~~~~~~~~~~
# Import the catalytic converter geometry file.

# Import geometry with platform-specific file handling

filenames = {
    "Windows": "catalytic_converter.dsco",
    "Other": "catalytic_converter.dsco.pmdb",
}

geometry_filename = examples.download_file(
    filenames.get(platform.system(), filenames["Other"]),
    "/pyfluent/catalytic_converter/",
    save_path=os.getcwd(),
)

workflow.InitializeWorkflow(WorkflowType="Watertight Geometry")
workflow.TaskObject["Import Geometry"].Arguments = dict(FileName=geometry_filename)
workflow.TaskObject["Import Geometry"].Execute()

# %%
# Local Sizing Controls
# ~~~~~~~~~~~~~~~~~~~~~
# Add local sizing controls for sensor components to ensure adequate mesh resolution.

# Add local sizing for sensor components

workflow.TaskObject["Add Local Sizing"].Arguments = dict(
    {
        "AddChild": "yes",
        "BOIControlName": "sensor",
        "BOIExecution": "Curvature",
        "BOIFaceLabelList": [
            "sensing_element-65-solid",
            "sensor_innertube-67-solid",
            "sensor_protectiontube-66-solid1",
        ],
        "BOIMaxSize": 1.2,
        "BOIMinSize": 0.1,
    }
)
workflow.TaskObject["Add Local Sizing"].AddChildToTask()
workflow.TaskObject["Add Local Sizing"].InsertCompoundChildTask()
workflow.TaskObject["Add Local Sizing"].Arguments = dict({"AddChild": "yes"})
workflow.TaskObject["sensor"].Execute()

# %%
# Surface Mesh Generation
# ~~~~~~~~~~~~~~~~~~~~~~~
# Generate surface mesh with quality improvements.

# Configure surface mesh settings

workflow.TaskObject["Generate the Surface Mesh"].Arguments = dict(
    {
        "CFDSurfaceMeshControls": {"MinSize": 1.5},
        "SurfaceMeshPreferences": {
            "SMQualityImprove": "yes",
            "SMQualityImproveLimit": 0.95,
            "ShowSurfaceMeshPreferences": True,
        },
    }
)
workflow.TaskObject["Generate the Surface Mesh"].Execute()

# %%
# Geometry Description
# ~~~~~~~~~~~~~~~~~~~~
# Describe the geometry setup with fluid and solid regions requiring capping.

# Describe geometry type

workflow.TaskObject["Describe Geometry"].UpdateChildTasks(SetupTypeChanged=False)
workflow.TaskObject["Describe Geometry"].Arguments = dict(
    {"SetupType": "The geometry consists of both fluid and solid regions and/or voids"}
)
workflow.TaskObject["Describe Geometry"].UpdateChildTasks(SetupTypeChanged=True)

# Enable capping and wall-to-internal conversion

workflow.TaskObject["Describe Geometry"].Arguments = dict(
    {
        "CappingRequired": "Yes",
        "SetupType": "The geometry consists of both fluid and solid regions and/or voids",
    }
)
workflow.TaskObject["Describe Geometry"].UpdateChildTasks(SetupTypeChanged=False)
workflow.TaskObject["Describe Geometry"].Arguments = dict(
    {
        "CappingRequired": "Yes",
        "SetupType": "The geometry consists of both fluid and solid regions and/or voids",
        "WallToInternal": "Yes",
    }
)
workflow.TaskObject["Describe Geometry"].Execute()

# %%
# Boundary Capping
# ~~~~~~~~~~~~~~~~
# Create inlet and outlet boundary patches through capping operations.

# Create inlet boundary

workflow.TaskObject["Enclose Fluid Regions (Capping)"].Arguments = dict(
    {
        "LabelSelectionList": ["in1"],
        "PatchName": "inlet",
    }
)
workflow.TaskObject["Enclose Fluid Regions (Capping)"].AddChildToTask()
workflow.TaskObject["Enclose Fluid Regions (Capping)"].InsertCompoundChildTask()
workflow.TaskObject["inlet"].Execute()

# Create outlet boundary as pressure outlet

workflow.TaskObject["Enclose Fluid Regions (Capping)"].Arguments = dict(
    {
        "LabelSelectionList": ["out1"],
        "PatchName": "outlet",
    }
)
workflow.TaskObject["Enclose Fluid Regions (Capping)"].AddChildToTask()
workflow.TaskObject["Enclose Fluid Regions (Capping)"].InsertCompoundChildTask()
workflow.TaskObject["outlet"].Execute()

# Configure outlet as pressure-outlet

workflow.TaskObject["outlet"].Revert()
workflow.TaskObject["outlet"].Arguments = dict(
    {
        "CompleteLabelSelectionList": ["out1"],
        "LabelSelectionList": ["out1"],
        "PatchName": "outlet",
        "ZoneType": "pressure-outlet",
    }
)
workflow.TaskObject["outlet"].Execute()

# Update boundaries

workflow.TaskObject["Update Boundaries"].Execute()

# %%
# Region Setup
# ~~~~~~~~~~~~
# Create and configure fluid regions including substrate conversion.

# Create multiple flow volumes

workflow.TaskObject["Create Regions"].Arguments = dict({"NumberOfFlowVolumes": 3})
workflow.TaskObject["Create Regions"].Execute()

# Convert solid substrate regions to fluid regions

workflow.TaskObject["Update Regions"].Arguments = dict(
    {
        "OldRegionNameList": ["honeycomb-solid1", "honeycomb_af0-solid1"],
        "OldRegionTypeList": ["solid", "solid"],
        "RegionNameList": ["fluid:substrate:1", "fluid:substrate:2"],
        "RegionTypeList": ["fluid", "fluid"],
    }
)
workflow.TaskObject["Update Regions"].Execute()

# %%
# Boundary Layer Mesh
# ~~~~~~~~~~~~~~~~~~~
# Add boundary layer mesh for accurate near-wall resolution.

# Add boundary layers

workflow.TaskObject["Add Boundary Layers"].AddChildToTask()
workflow.TaskObject["Add Boundary Layers"].InsertCompoundChildTask()
workflow.TaskObject["smooth-transition_1"].Arguments = dict(
    {"BLControlName": "smooth-transition_1"}
)

workflow.TaskObject["smooth-transition_1"].Execute()

# %%
# Volume Mesh Generation
# ~~~~~~~~~~~~~~~~~~~~~~
# Generate the final volume mesh with body label merging.

# Generate volume mesh

workflow.TaskObject["Generate the Volume Mesh"].Arguments = dict(
    {"VolumeMeshPreferences": {"MergeBodyLabels": "yes"}}
)
workflow.TaskObject["Generate the Volume Mesh"].Execute()

# %%
# Mesh Quality Check & Write Mesh File
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Check mesh quality and write mesh files and snapshot.

# Check mesh quality

meshing.tui.mesh.check_mesh()

# Write mesh files

meshing.tui.file.write_mesh("out/catalytic_converter.msh.h5")
# %%
# Switch to Solver
# ^^^^^^^^^^^^^^^^
# Switch from meshing to solver mode for setup and solution.

# Switch to solver mode

solver_session = meshing.switch_to_solver()

# %% Visualize Mesh
# ~~~~~~~~~~~~~~~~~

# Visualize mesh in the graphics window and save a screenshot
# Create Graphics object to save the graphics with following settings

graphics = Graphics(solver_session)
graphics.views.auto_scale()
if graphics.picture.use_window_resolution.is_active():
    graphics.picture.use_window_resolution = False

graphics.views.restore_view(view_name="isometric")
graphics.picture.x_resolution = 600
graphics.picture.y_resolution = 600
graphics.picture.color_mode = "color"

# Define and display the mesh
# First, get all wall boundary names and create a mesh object for visualization context

all_walls = solver_session.settings.setup.boundary_conditions.wall.get_object_names()
mesh = Mesh(solver_session, new_instance_name="mesh-1")
mesh.surfaces_list = all_walls
mesh.options.edges = True
mesh.display()
os.makedirs("out", exist_ok=True)
graphics.picture.save_picture(file_name="out/catalytic_converter_mesh.png")
mesh.options.edges = False  # Turn off edges after saving the picture

# %%
# .. figure:: /_static/catalytic_converter/catalytic_converter_mesh.png
#     :width: 500pt
#     :align: center
#
#     Polyhedra mesh for the catalytic converter geometry.

# %%
# General Setup
# ~~~~~~~~~~~~~
# Configure units and enable energy equation.

# Set length units to millimeters

general_settings = General(solver_session)
general_settings.units_settings.new_unit(
    offset=0.0, units_name="mm", scale_factor=1.0, quantity="length"
)
# Enable energy equation
energy_model_settings = Models(solver_session)
energy_model_settings.energy = {
    "enabled": True,
    "inlet_diffusion": False,
}

# %%
# Materials
# ~~~~~~~~~
# Set up material properties for the fluid.

# Copy nitrogen from database

solver_session.settings.setup.materials.database.copy_by_name(
    type="fluid", name="nitrogen"
)

# Assign material to main fluid zone

materials = Materials(solver_session)
materials.database.copy_by_name(type="fluid", name="nitrogen")

material_assignments = CellZoneCondition(solver_session, name="fluid:1")
material_assignments.general = {"material": "nitrogen"}

# Copy boundary conditions to other fluid zones

solver_session.tui.define.boundary_conditions.copy_bc(
    "fluid:1", "fluid:2", "fluid:3", "()"
)

# %%
# Cell Zone Conditions
# ~~~~~~~~~~~~~~~~~~~~
# Configure porous media properties for catalytic substrate zones.

# Configure first substrate zone as porous media

porous_media_settings = CellZoneCondition(solver_session, name="fluid:substrate:1")
porous_media_settings.general = {"laminar": True, "material": "nitrogen"}
porous_media_settings.porous_zone = {
    "solid_material": "aluminum",
    "equib_thermal": True,
    "relative_viscosity": {"value": 1, "option": "constant"},
    "porosity": 1,
    "power_law_model": [0, 0],
    "inertial_resistance": [1000.0, 1000.0, 1000.0],
    "alt_inertial_form": False,
    "viscous_resistance": [1000000.0, 1000000.0, 1000.0],
    "rel_vel_resistance": True,
    "direction_2_vector": [0, 1, 0],
    "direction_1_vector": [1, 0, 0],
    "dir_spec_cond": "Cartesian",
    "porous": True,
}


# Copy porous media settings to second substrate zone

solver_session.tui.define.boundary_conditions.copy_bc(
    "fluid:substrate:1", "fluid:substrate:2", "()"
)

# %%
# Boundary Conditions
# ~~~~~~~~~~~~~~~~~~~
# Set up inlet and outlet boundary conditions.

# Configure velocity inlet

velocity_inlet = VelocityInlet(solver_session, name="inlet")
velocity_inlet.momentum = {"velocity_magnitude": {"value": 125.0}}

velocity_inlet.turbulence = {
    "hydraulic_diameter": 0.5,  # in meters
    "turbulence_specification": "Intensity and Hydraulic Diameter",
}
velocity_inlet.thermal = {"temperature": {"value": 800.0}}  # in Kelvin

# Configure pressure outlet

pressure_outlet = PressureOutlet(solver_session, name="outlet")
pressure_outlet.momentum = {"gauge_pressure": {"value": 0.0}}
pressure_outlet.turbulence = {
    "backflow_hydraulic_diameter": 0.5,  # in meters
    "turbulence_specification": "Intensity and Hydraulic Diameter",
}
pressure_outlet.thermal = {"backflow_total_temperature": {"value": 800.0}}  # in Kelvin


# %%
# Solution Monitoring
# ~~~~~~~~~~~~~~~~~~~
# Set up surface monitors for mass flow rate tracking.

# Create surface report definition

report_definitions = ReportDefinitions(solver_session)
surface_report = report_definitions.surface.create(name="surf-mon-1")
surface_report.report_type = "surface-massflowrate"
surface_report.surface_names = ["outlet"]

# Create report file monitor

monitor = Monitor(solver_session)
surface_monitor = monitor.report_files.create(name="surf-mon-1")
surface_monitor.print = True
surface_monitor.report_defs = ["surf-mon-1"]
os.makedirs("out", exist_ok=True)
surface_monitor.file_name = "out/surf-mon-1.out"

# Create report plot monitor

surface_plot_monitor = monitor.report_plots.create(name="surf-mon-1")
surface_plot_monitor.print = True
surface_plot_monitor.report_defs = ["surf-mon-1"]


# %%
# Initialization
# ~~~~~~~~~~~~~~
# Initialize the flow field.

# Compute initial conditions from inlet

solution_initialization = Initialization(solver_session)

solution_initialization.compute_defaults(
    from_zone_type="velocity-inlet", from_zone_name="inlet", phase="mixture"
)

# Set initialization method and initialize

solution_initialization.initialization_type = "standard"
solution_initialization.standard_initialize()

# %%
# Run Calculation
# ~~~~~~~~~~~~~~~
# Execute the iterative solution process.

# Set iteration count and run calculation

run_calculation = RunCalculation(solver_session)
run_calculation.iter_count = (
    150  # Iteration count, keep it at 150 for demo purposes only
)
run_calculation.calculate()


# %%
# Results Processing
# ^^^^^^^^^^^^^^^^^^
# Extract and analyze simulation results.

# Mass Flow Analysis
# ~~~~~~~~~~~~~~~~~~
# Calculate mass flow rate at outlet.

# Ensure the 'out/' directory exists before writing files to it
os.makedirs("out", exist_ok=True)
results = Results(solver_session)
results.report.fluxes.mass_flow(
    zones=["outlet"], write_to_file=True, file_name="out/mass_flow_rate.flp"
)

# %%
# Surface Creation
# ~~~~~~~~~~~~~~~~
# Create iso-surfaces for flow field visualization.

# Create cross-sectional surfaces

surfaces_data = [
    ("y=-425", "y-coordinate", [-0.425]),
    ("z=185", "z-coordinate", [0.185]),
    ("z=230", "z-coordinate", [0.23]),
    ("z=280", "z-coordinate", [0.28]),
    ("z=330", "z-coordinate", [0.33]),
    ("z=375", "z-coordinate", [0.375]),
]

for surf_name, field_name, iso_values in surfaces_data:
    results.surfaces.iso_surface.create(name=surf_name)
    results.surfaces.iso_surface[surf_name].field = field_name
    results.surfaces.iso_surface[surf_name].iso_values = iso_values

# %%
# Velocity Analysis
# ~~~~~~~~~~~~~~~~~
# Calculate mass-weighted average velocity at different axial locations.

z_surfaces = ["z=185", "z=230", "z=280", "z=330", "z=375"]
results.report.surface_integrals.mass_weighted_avg(
    surface_names=z_surfaces,
    report_of="velocity-magnitude",
    write_to_file=True,
    file_name="out/mass-avg_vel.srp",
)

# %%
# Vector Plot
# ~~~~~~~~~~~
# Create velocity vector, static pressure, and velocity magnitude plots.
#
# This section demonstrates post-processing visualization techniques using PyFluent's
# graphics capabilities. We'll create three different types of visualizations:
#
# 1. **Velocity Vectors**: Show flow direction and magnitude using vectors
# 2. **Static Pressure Contours**: Display pressure distribution along a cross-section
# 3. **Velocity Magnitude Contours**: Visualize velocity distribution across multiple surfaces
#
# Each visualization is displayed in a separate scene with consistent mesh transparency
# to provide context and better understanding of the flow field.

# Create velocity vector plot
# Velocity vectors show both flow direction and magnitude using arrow symbols
# The scale factor controls arrow size for optimal visualization

results.graphics.vector.create(name="vector-vel")
results.graphics.vector["vector-vel"] = {
    "scale": {
        "auto_scale": True,  # Automatically scale arrows for best visibility
        "scale_f": 0.006,  # Fine-tune scale factor (smaller = shorter arrows)
    },
    "surfaces_list": ["y=-425"],  # Display vectors on the y=-425 cross-section
}

# Create static pressure contour plot
# Pressure contours use color mapping to show pressure distribution
# This helps identify high/low pressure regions and pressure gradients

results.graphics.contour.create(name="contour-ps")
results.graphics.contour["contour-ps"] = {
    "field": "pressure",  # Specify pressure as the contour variable
    "surfaces_list": ["y=-425"],  # Display on the same cross-section as vectors
}

# Create velocity magnitude contour plot
# Velocity magnitude shows speed distribution without directional information
# Using multiple z-surfaces provides a comprehensive view through the domain

results.graphics.contour.create(name="contour-velmag")
results.graphics.contour["contour-velmag"] = {
    "field": "velocity-magnitude",  # Specify velocity magnitude as the variable
    "surfaces_list": z_surfaces,  # Display on all z-coordinate surfaces
}

# Scene 1: Display velocity vectors with mesh context
# Scenes combine multiple graphics objects for comprehensive visualization

velocity_vectors_scene = Scene(solver_session, new_instance_name="scene-1")
velocity_vectors_scene.graphics_objects.add(name="vector-vel")

# adding mesh for context which is created earlier steps just after switching to solver

velocity_vectors_scene.graphics_objects.add(name="mesh-1")

# Configure scene appearance and display settings
solver_session.settings.results.scene["scene-1"] = {
    "graphics_objects": {
        "vector-vel": {"name": "vectors"},  # Label for the vector plot
        "mesh-1": {"transparency": 75},  # Semi-transparent mesh (75% transparent)
    }
}
velocity_vectors_scene.display()
graphics.views.auto_scale()
os.makedirs("out", exist_ok=True)
graphics.picture.save_picture(file_name="out/velocity_vectors.png")

# %%
# .. figure:: /_static/catalytic_converter/velocity_vectors.png
#     :width: 500pt
#     :align: center
#
#     Velocity Vectors Through the Interior

# Scene 2: Display static pressure contours with mesh context
# This scene focuses on pressure distribution across the catalytic converter

static_pressure_scene = Scene(solver_session, new_instance_name="scene-2")
static_pressure_scene.graphics_objects.add(name="contour-ps")
static_pressure_scene.graphics_objects.add(name="mesh-1")

# Configure pressure contour scene settings

solver_session.settings.results.scene["scene-2"] = {
    "graphics_objects": {
        "contour-ps": {"name": "contour-ps"},  # Label for the pressure contour
        "mesh-1": {"transparency": 75},  # Consistent mesh transparency
    }
}
static_pressure_scene.display()
graphics.views.auto_scale()
# Ensure "out/" directory exists before saving files
os.makedirs("out", exist_ok=True)
graphics.picture.save_picture(file_name="out/static_pressure.png")

# %%
# .. figure:: /_static/catalytic_converter/static_pressure.png
#     :width: 500pt
#     :align: center
#
#     Contours of Static Pressure Through the Interior


# Scene 3: Display velocity magnitude contours with mesh context
# This scene shows speed distribution across multiple axial locations

velocity_magnitude_scene = Scene(solver_session, new_instance_name="scene-3")
velocity_magnitude_scene.graphics_objects.add(name="contour-velmag")
velocity_magnitude_scene.graphics_objects.add(name="mesh-1")

# Configure velocity magnitude contour scene settings
solver_session.settings.results.scene["scene-3"] = {
    "graphics_objects": {
        "contour-velmag": {
            "name": "contour-velmag"
        },  # Label for velocity magnitude contour
        "mesh-1": {"transparency": 75},  # Consistent mesh transparency
    }
}
velocity_magnitude_scene.display()
graphics.views.auto_scale()
# Ensure the 'out/' directory exists before saving files
os.makedirs("out", exist_ok=True)
graphics.picture.save_picture(file_name="out/velocity_magnitude.png")

# %%
# .. figure:: /_static/catalytic_converter/velocity_magnitude.png
#     :width: 500pt
#     :align: center
#
#     Contours of Velocity Magnitude on the z=185, z=230, z=280, z=330, and z=375 Surfaces


# %%
# File Output
# ^^^^^^^^^^^
# Save final case and data files.

# Write final case and data

solver_session.settings.file.write(
    file_type="case-data", file_name="out/catalytic_converter_final.cas.h5"
)

# %%
# Solver Exit
# ^^^^^^^^^^^
# Display simulation statistics and completion message.

# Close solver session

solver_session.exit()

# %%
# References
# ----------
# .. _References_1:
# .. [1] `ANSYS Fluent User's Guide, ANSYS, Inc. <https://ansyshelp.ansys.com/public/account/secured?returnurl=/Views/Secured/prod_page.html?pn=Fluent&prodver=25.2&lang=en>`_.
# .. _References_2:
# .. [2] `Modeling Flow Through Porous Media <https://ansyshelp.ansys.com/public/account/secured?returnurl=/Views/Secured/corp/v252/en/flu_tg/flu_tg_tut_catalytic_converter.html>`_.



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import warnings

import pytest

from ansys.fluent.core import PyFluentDeprecationWarning
from ansys.fluent.core.utils.deprecate import deprecate_arguments, deprecate_function


def test_deprecate_argument_warns():
    """Test that deprecated arguments triggers a warning and maps correctly."""

    @deprecate_arguments(old_args=["a", "b"], new_args=["d", "e"], version="3.0.0")
    @deprecate_arguments(old_args="c", new_args="f", version="3.0.0")
    @deprecate_arguments(old_args="x1", new_args="z1", version="3.0.0")
    def example_func(**kwargs):
        assert "d" in kwargs
        assert "f" in kwargs
        assert "z1" in kwargs
        assert "a" not in kwargs
        assert "c" not in kwargs
        assert "x1" not in kwargs
        return kwargs

    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")  # ensure deprecation warnings are captured
        result = example_func(a=1, b=2, c=3, x1=2, x2=4)

    # Check that the warnings were raised
    assert len(w) == 6
    warning = w[0]
    assert issubclass(warning.category, PyFluentDeprecationWarning)
    assert "'a', 'b'" in str(warning.message)
    assert "'d', 'e'" in str(warning.message)

    # Check that function received the new argument
    assert result == {"d": 1, "e": 2, "f": 3, "x2": 4, "z1": 2}


def test_deprecate_argument_warns_with_custom_converter():
    """Test that deprecated arguments triggers a warning and maps correctly."""

    def my_custom_converter(kwargs, old_args, new_args):
        # Combine values of a, b into a tuple → d
        values = tuple(kwargs.pop(arg) for arg in old_args if arg in kwargs)
        if values:
            kwargs[new_args[0]] = values
        return kwargs

    @deprecate_arguments(
        old_args=["a", "b"],
        new_args="d",
        version="3.0.0",
        converter=my_custom_converter,
    )
    @deprecate_arguments(
        old_args="c",
        new_args="e",
        version="3.0.0",
    )
    def example_func(**kwargs):
        assert "d" in kwargs
        assert "e" in kwargs
        assert "c" not in kwargs
        assert "a" not in kwargs
        return kwargs

    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")  # ensure deprecation warnings are captured
        result = example_func(a=1, b=2, c=3)

    # Check that the warnings were raised
    assert len(w) == 4
    warning = w[0]
    assert issubclass(warning.category, PyFluentDeprecationWarning)
    assert "'a', 'b'" in str(warning.message)
    assert "'d'" in str(warning.message)

    assert result == {"d": (1, 2), "e": 3}


def test_deprecate_argument_raises_value_error_without_converter():
    """Test that deprecated arguments raises ValueError for incorrect mapping."""

    with pytest.raises(ValueError):

        @deprecate_arguments(
            old_args=["a", "b"],
            new_args="d",
            version="3.0.0",
        )
        def example_func_1(**kwargs):
            assert "d" in kwargs
            assert "e" in kwargs
            assert "c" not in kwargs
            assert "a" not in kwargs
            return kwargs

    with pytest.raises(ValueError):

        @deprecate_arguments(
            old_args="a",
            new_args=["d", "e"],
            version="3.0.0",
        )
        def example_func_2(**kwargs):
            assert "d" in kwargs
            assert "e" in kwargs
            assert "c" not in kwargs
            assert "a" not in kwargs
            return kwargs


def test_deprecate_function():
    """Test that deprecated function display warning."""

    @deprecate_function(version="3.0.0", new_func="new_add")
    def old_add(a, b):
        return a + b

    def new_add(a, b):
        return a + b + 10

    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")  # ensure deprecation warnings are captured
        result = old_add(1, 2)

    # Check that the warnings were raised
    assert len(w) == 2
    warning = w[0]
    assert issubclass(warning.category, PyFluentDeprecationWarning)
    assert "'old_add'" in str(warning.message)
    assert "'new_add'" in str(warning.message)
    assert "3.0.0" in str(warning.message)

    assert result == 3



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides a module for launching and configuring local Fluent Docker container runs.

Notes
-----

For configuration details, see :func:`configure_container_dict`, and for a list of additional Docker container run
configuration options that can also be specified through the
``container_dict`` argument for :func:`~ansys.fluent.core.launcher.launcher.launch_fluent()`,
see documentation for `Docker run`_.

For the Fluent Docker container to be able to find license information, the license file or server needs to be specified
through the ``ANSYSLMD_LICENSE_FILE`` environment variable,
or the ``license_server`` argument for the ``container_dict`` (see :func:`configure_container_dict`).

.. _Docker run: https://docker-py.readthedocs.io/en/stable/containers.html#docker.models.containers.ContainerCollection.run

Examples
--------
Launching a Fluent Docker container with system default configuration:

>>> import ansys.fluent.core as pyfluent
>>> session = pyfluent.launch_fluent(start_container=True)

Launching with custom configuration, using ``mount_source`` and ``fluent_image``
which are arguments for :func:`configure_container_dict`, and ``auto_remove`` which is an argument for `Docker run`_:

>>> import ansys.fluent.core as pyfluent
>>> custom_config = {}
>>> custom_config.update(fluent_image='custom_fluent:v23.1.0', mount_source='/testing', auto_remove=False)
>>> session = pyfluent.launch_fluent(container_dict=custom_config)

Getting default Fluent Docker container configuration, then launching with customized configuration:

>>> import ansys.fluent.core as pyfluent
>>> config_dict = pyfluent.launch_fluent(start_container=True, dry_run=True)
Docker container run configuration information:
config_dict =
{'auto_remove': True,
 'command': ['-gu', '-sifile=/mnt/pyfluent/serverinfo-lpqsdldw.txt', '3ddp'],
 'detach': True,
 'environment': {'ANSYSLMD_LICENSE_FILE': '2048@licenseserver.com',
                 'REMOTING_PORTS': '54000/portspan=2'},
 'fluent_image': '<image registry>:v23.2.0',
 'labels': {'test_name': 'none'},
 'ports': {'54000': 54000},
 'volumes': ['/home/user/.local/share/ansys_fluent_core/examples:/mnt/pyfluent'],
 'working_dir': '/mnt/pyfluent'}
>>> config_dict.update(image_name='custom_fluent', image_tag='v23.1.0', mem_limit='1g')
>>> session = pyfluent.launch_fluent(container_dict=config_dict)
"""

import logging
import os
from pathlib import Path, PurePosixPath
from pprint import pformat
import tempfile
from typing import Any, List
import warnings

import ansys.fluent.core as pyfluent
from ansys.fluent.core.docker.docker_compose import ComposeBasedLauncher
from ansys.fluent.core.docker.utils import get_ghcr_fluent_image_name
from ansys.fluent.core.launcher.error_handler import (
    LaunchFluentError,
)
from ansys.fluent.core.launcher.launcher_utils import ComposeConfig
from ansys.fluent.core.pyfluent_warnings import PyFluentDeprecationWarning
from ansys.fluent.core.session import _parse_server_info_file
from ansys.fluent.core.utils.deprecate import deprecate_arguments
from ansys.fluent.core.utils.execution import timeout_loop
from ansys.fluent.core.utils.networking import get_free_port

logger = logging.getLogger("pyfluent.launcher")


class FluentImageNameTagNotSpecified(ValueError):
    """Raised when Fluent image name or image tag is not specified."""

    def __init__(self):
        """Initializes FluentImageNameTagNotSpecified."""
        super().__init__(
            "Specify either 'fluent_image' or 'image_tag' and 'image_name'."
        )


class ServerInfoFileError(ValueError):
    """Raised when server info file is not given properly."""

    def __init__(self):
        """Initializes ServerInfoFileError."""
        super().__init__(
            "Specify server info file either using 'container_server_info_file' argument or in the 'container_dict'."
        )


class LicenseServerNotSpecified(KeyError):
    """Raised when license server is not specified."""

    def __init__(self):
        """Initializes LicenseServerNotSpecified."""
        super().__init__(
            "Specify licence server either using 'ANSYSLMD_LICENSE_FILE' environment variable or in the 'container_dict'."
        )


def dict_to_str(dict: dict) -> str:
    """Converts the dict to string while hiding the 'environment' argument from the dictionary,
    if the environment variable 'PYFLUENT_HIDE_LOG_SECRETS' is '1'.
    This is useful for logging purposes, to avoid printing sensitive information such as license server details.
    """

    if "environment" in dict and pyfluent.config.hide_log_secrets:
        modified_dict = dict.copy()
        modified_dict.pop("environment")
        return pformat(modified_dict)
    else:
        return pformat(dict)


@deprecate_arguments(
    old_args="container_mount_path",
    new_args="mount_target",
    version="v0.23.0",
)
@deprecate_arguments(
    old_args="host_mount_path",
    new_args="mount_source",
    version="v0.23.0",
)
def configure_container_dict(
    args: List[str],
    mount_source: str | Path | None = None,
    mount_target: str | Path | None = None,
    timeout: int | None = None,
    port: int | None = None,
    license_server: str | None = None,
    container_server_info_file: str | Path | None = None,
    remove_server_info_file: bool = True,
    fluent_image: str | None = None,
    image_name: str | None = None,
    image_tag: str | None = None,
    file_transfer_service: Any | None = None,
    compose_config: ComposeConfig | None = None,
    certificates_folder: str | None = None,
    **container_dict,
) -> (dict, int, int, Path, str, bool):
    """Parses the parameters listed below, and sets up the container configuration file.

    Parameters
    ----------
    args : List[str]
        List of Fluent launch arguments.
    mount_source : str | Path, optional
        Path on the host system to mount into the container. This directory will serve as the working directory
        for the Fluent process inside the container. If not specified, PyFluent's current working directory will
        be used.
    mount_target : str | Path, optional
        Path inside the container where ``mount_source`` will be mounted. This will be the working directory path
        visible to the Fluent process running inside the container.
    timeout : int, optional
        Time limit for the Fluent container to start, in seconds.
    port : int, optional
        Port for Fluent container to use.
    license_server : str, optional
        License server for Ansys Fluent to use.
    container_server_info_file : str | Path, optional
        Name of the server information file for Fluent to write on the ``mount_source``.
    remove_server_info_file : bool, optional
        Defaults to True, and automatically deletes the server information file after PyFluent has finished using it.
    fluent_image : str, optional
        Specifies full image name for Docker container run, with the format ``"image_name:image_tag"``.
        ``image_tag`` and ``image_name`` are ignored if ``fluent_image`` has been specified.
    image_name : str, optional
        Ignored if ``fluent_image`` has been specified.
    image_tag : str, optional
        Ignored if ``fluent_image`` has been specified.
    file_transfer_service : optional
        Supports file upload and download.
    compose_config : ComposeConfig, optional
        Configuration for Docker Compose, if using Docker Compose to launch the container.
    certificates_folder : str, optional
        Path to the folder containing TLS certificates for Fluent's gRPC server.
    **container_dict
        Additional keyword arguments can be specified, they will be treated as Docker container run options
        to be passed directly to the Docker run execution. See examples below and `Docker run`_ documentation.

    Returns
    -------
    fluent_image : str
    container_dict : dict
    timeout : int
    port : int
    host_server_info_file : Path
    container_server_info_file: str
    remove_server_info_file: bool

    Raises
    ------
    LicenseServerNotSpecified
        If license server is not specified through an environment variable or in ``container_dict``.
    ServerInfoFileError
        If server info file is specified through both a command-line argument inside ``container_dict`` and the  ``container_server_info_file`` parameter.
    FluentImageNameTagNotSpecified
        If ``fluent_image`` or ``image_tag`` and ``image_name`` are not specified.

    Notes
    -----
    This function should usually not be called directly, it is automatically used by
    :func:`~ansys.fluent.core.launcher.launcher.launch_fluent()`.

    For a list of additional Docker container run configuration options that can also be specified using
    ``container_dict``, see `Docker run`_ documentation.

    See also :func:`start_fluent_container`.
    """

    compose_config = compose_config if compose_config else ComposeConfig()

    if timeout is not None:
        warnings.warn(
            "configure_container_dict(timeout) is deprecated, use launch_fluent(start_timeout) instead.",
            PyFluentDeprecationWarning,
        )

    logger.debug(f"container_dict before processing:\n{dict_to_str(container_dict)}")

    # Starting with 'mount_source' because it is not tied to the 'working_dir'.
    # The intended 'mount_source' logic is as follows, if it is not directly specified:
    # 1. If 'file_transfer_service' is provided, use its 'mount_source'.
    # 2. Use the value from 'pyfluent.config.container_mount_source', if it is set.
    # 3. If 'volumes' is specified in 'container_dict', try to infer the value from it.
    # 4. Finally, use the current working directory, which is always available.

    if not mount_source:
        if file_transfer_service:
            mount_source = file_transfer_service.mount_source
        else:
            mount_source = pyfluent.config.container_mount_source

    if "volumes" in container_dict:
        if len(container_dict["volumes"]) != 1:
            logger.warning(
                "Multiple volumes being mounted in the Docker container, "
                "Assuming the first mount is the working directory for Fluent."
            )
        volumes_string = container_dict["volumes"][0]
        if mount_source:
            logger.warning(
                "'volumes' keyword specified in 'container_dict', but "
                "it is going to be overwritten by specified 'mount_source'."
            )
        else:
            mount_source = volumes_string.split(":")[0]
            logger.debug(f"mount_source: {mount_source}")
        inferred_mount_target = volumes_string.split(":")[1]
        logger.debug(f"inferred_mount_target: {inferred_mount_target}")

    if not mount_source:
        logger.debug("No container 'mount_source' specified, using default value.")
        mount_source = os.getcwd()

    # The intended 'mount_target' logic is as follows, if it is not directly specified:
    # 1. If 'working_dir' is specified in 'container_dict', use it as 'mount_target'.
    # 2. Try to infer the value from the 'volumes' keyword in 'container_dict', if available.
    # 3. Finally, use the value from 'pyfluent.config.container_mount_target', which is always set.

    if not mount_target:
        if "working_dir" in container_dict:
            mount_target = container_dict["working_dir"]
        else:
            mount_target = pyfluent.config.container_mount_target

    if "working_dir" in container_dict and mount_target:
        # working_dir will be set later to the final value of mount_target
        container_dict.pop("working_dir")

    if not mount_target and "volumes" in container_dict:
        mount_target = inferred_mount_target

    if not mount_target:
        logger.debug("No container 'mount_target' specified, using default value.")
        mount_target = pyfluent.config.container_mount_target

    if "volumes" not in container_dict:
        container_dict.update(volumes=[f"{mount_source}:{mount_target}"])
    else:
        container_dict["volumes"][0] = f"{mount_source}:{mount_target}"

    logger.warning(
        f"Configuring Fluent container to mount to {mount_source}, "
        f"with this path available as {mount_target} for the Fluent session running inside the container."
    )

    if "working_dir" not in container_dict:
        container_dict.update(
            working_dir=mount_target,
        )

    if certificates_folder is not None:
        certs_mount = f"{certificates_folder}:/tmp/certs:ro"
        if certs_mount not in container_dict["volumes"]:
            container_dict["volumes"].append(certs_mount)
    port_mapping = {port: port} if port else {}
    if not port_mapping and "ports" in container_dict:
        # take the specified 'port', OR the first port value from the specified 'ports', for Fluent to use
        port_mapping = container_dict["ports"]
    if not port_mapping and pyfluent.config.launch_fluent_port:
        port = pyfluent.config.launch_fluent_port
        port_mapping = {port: port}
    if not port_mapping:
        port = get_free_port()
        port_mapping = {port: port}

    container_dict.update(
        ports={str(x): y for x, y in port_mapping.items()}
    )  # container port : host port
    container_grpc_port = next(
        iter(port_mapping.values())
    )  # the first port in the mapping is chosen as the gRPC port

    if "environment" not in container_dict:
        if not license_server:
            license_server = os.getenv("ANSYSLMD_LICENSE_FILE")

        if not license_server:
            raise LicenseServerNotSpecified()
        container_dict.update(
            environment={
                "ANSYSLMD_LICENSE_FILE": license_server,
                "REMOTING_PORTS": f"{container_grpc_port}/portspan=2",
                "FLUENT_ALLOW_REMOTE_GRPC_CONNECTION": "1",
            }
        )

    if "labels" not in container_dict:
        test_name = pyfluent.config.test_name
        container_dict.update(
            labels={"test_name": test_name},
        )

    # Find the server info file name from the command line arguments
    if "command" in container_dict:
        for v in container_dict["command"]:
            if v.startswith("-sifile="):
                if container_server_info_file:
                    raise ServerInfoFileError()
                container_server_info_file = PurePosixPath(
                    v.replace("-sifile=", "")
                ).name
                logger.debug(
                    f"Found server info file specification for {container_server_info_file}."
                )

    if container_server_info_file:
        container_server_info_file = (
            PurePosixPath(mount_target) / PurePosixPath(container_server_info_file).name
        )
    else:
        fd, sifile = tempfile.mkstemp(
            suffix=".txt", prefix="serverinfo-", dir=mount_source
        )
        os.close(fd)
        container_server_info_file = PurePosixPath(mount_target) / Path(sifile).name

    logger.debug(
        f"Using server info file '{container_server_info_file}' for Fluent container."
    )

    # If the 'command' had already been specified in the 'container_dict',
    # maintain other 'command' arguments but update the '-sifile' argument,
    # as the 'mount_target' or 'working_dir' may have changed.
    if "command" in container_dict:
        for i, item in enumerate(container_dict["command"]):
            if item.startswith("-sifile="):
                container_dict["command"][i] = f"-sifile={container_server_info_file}"
    else:
        container_dict["command"] = args + [f"-sifile={container_server_info_file}"]

    if not fluent_image:
        if not image_tag:
            image_tag = pyfluent.config.fluent_image_tag
        if not image_name and image_tag:
            image_name = (
                pyfluent.config.fluent_image_name
                or get_ghcr_fluent_image_name(image_tag)
            )
        if not image_tag or not image_name:
            fluent_image = pyfluent.config.fluent_container_name
        elif image_tag and image_name:
            if image_tag.startswith("sha"):
                fluent_image = f"{image_name}@{image_tag}"
            else:
                fluent_image = f"{image_name}:{image_tag}"
        else:
            raise FluentImageNameTagNotSpecified()

    container_dict["fluent_image"] = fluent_image

    if not pyfluent.config.fluent_automatic_transcript:
        if "environment" not in container_dict:
            container_dict["environment"] = {}
        container_dict["environment"]["FLUENT_NO_AUTOMATIC_TRANSCRIPT"] = "1"

    if pyfluent.config.launch_fluent_ip or pyfluent.config.remoting_server_address:
        if "environment" not in container_dict:
            container_dict["environment"] = {}
        container_dict["environment"]["REMOTING_SERVER_ADDRESS"] = (
            pyfluent.config.launch_fluent_ip or pyfluent.config.remoting_server_address
        )

    if pyfluent.config.launch_fluent_skip_password_check:
        if "environment" not in container_dict:
            container_dict["environment"] = {}
        container_dict["environment"]["FLUENT_LAUNCHED_FROM_PYFLUENT"] = "1"

    container_dict_base = {}
    container_dict_base.update(
        detach=True,
        auto_remove=True,
    )

    for k, v in container_dict_base.items():
        if k not in container_dict:
            container_dict[k] = v

    if not Path(mount_source).exists():
        Path(mount_source).mkdir(parents=True, exist_ok=True)
    host_server_info_file = Path(mount_source) / container_server_info_file.name

    if compose_config.is_compose:
        container_dict["host_server_info_file"] = host_server_info_file
        container_dict["mount_source"] = mount_source
        container_dict["mount_target"] = mount_target

    logger.debug(
        f"Fluent container container_grpc_port: {container_grpc_port}, "
        f"host_server_info_file: '{host_server_info_file}', "
        f"remove_server_info_file: {remove_server_info_file}"
    )
    logger.debug(f"container_dict after processing:\n{dict_to_str(container_dict)}")

    return (
        container_dict,
        timeout,
        container_grpc_port,
        host_server_info_file,
        container_server_info_file,
        remove_server_info_file,
    )


def start_fluent_container(
    args: List[str],
    container_dict: dict | None = None,
    start_timeout: int = 60,
    compose_config: ComposeConfig | None = None,
) -> tuple[int, str, Any]:
    """Start a Fluent container.

    Parameters
    ----------
    args : List[str]
        List of Fluent launch arguments.
    container_dict : dict, optional
        Dictionary with Docker container configuration.
    start_timeout : int, optional
        Timeout in seconds for the container to start. If not specified, it defaults to 60
        seconds.
    compose_config : ComposeConfig, optional
        Configuration for Docker Compose, if using Docker Compose to launch the container.

    Returns
    -------
    int
        Fluent gPRC server port exposed from the container.
    str
        Fluent gPRC server password exposed from the container.

    Raises
    ------
    TimeoutError
        If Fluent container launch reaches timeout.

    Notes
    -----
    Uses :func:`configure_container_dict` to parse the optional ``container_dict`` configuration.

    This function should usually not be called directly, it is automatically used by
    :func:`~ansys.fluent.core.launcher.launcher.launch_fluent()`.
    """

    compose_config = compose_config if compose_config else ComposeConfig()

    if container_dict is None:
        container_dict = {}

    container_vars = configure_container_dict(
        args,
        compose_config=compose_config,
        **container_dict,
    )

    (
        config_dict,
        timeout,
        port,
        host_server_info_file,
        container_server_info_file,
        remove_server_info_file,
    ) = container_vars
    launch_string = " ".join(config_dict["command"])

    if timeout:
        logger.warning(
            "launch_fluent(start_timeout) overridden by configure_container_dict(timeout) value."
        )
        start_timeout = timeout
        del timeout

    try:
        if compose_config.is_compose:
            config_dict["fluent_port"] = port

            compose_container = ComposeBasedLauncher(
                compose_config=compose_config,
                container_dict=config_dict,
                container_server_info_file=container_server_info_file,
            )

            if not compose_container.check_image_exists():
                logger.debug(
                    f"Fluent image {config_dict['fluent_image']} not found. Pulling image..."
                )
                compose_container.pull_image()

            compose_container.start()

            return port, config_dict, compose_container
        else:
            if not host_server_info_file.exists():
                host_server_info_file.parents[0].mkdir(exist_ok=True)

            host_server_info_file.touch(exist_ok=True)
            last_mtime = host_server_info_file.stat().st_mtime

            import docker

            docker_client = docker.from_env()

            logger.debug("Starting Fluent docker container...")

            container = docker_client.containers.run(
                config_dict.pop("fluent_image"), **config_dict
            )

            logger.debug(
                f"Waiting for Fluent container for up to {start_timeout} seconds..."
            )

            success = timeout_loop(
                lambda: host_server_info_file.stat().st_mtime > last_mtime,
                start_timeout,
            )

            if not success:
                try:
                    container.stop()
                except Exception as stop_ex:
                    logger.error(f"Failed to stop container: {stop_ex}")
                    raise TimeoutError(
                        f"Fluent container launch has timed out after {start_timeout} seconds. "
                        f"Additionally, stopping the container failed: {stop_ex}"
                    ) from stop_ex
                else:
                    raise TimeoutError(
                        f"Fluent container launch has timed out after {start_timeout} seconds."
                        " The container was stopped."
                    )
            else:
                _, _, password = _parse_server_info_file(str(host_server_info_file))

                return port, password, container
    except Exception as ex:
        logger.error(f"Exception caught - {type(ex).__name__}: {ex}")
        raise LaunchFluentError(launch_string) from ex
    finally:
        if remove_server_info_file and host_server_info_file.exists():
            host_server_info_file.unlink()



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides a module to get warnings for core functionality."""

import warnings


class PyFluentDeprecationWarning(FutureWarning):
    """Provides the common warning class for warnings about deprecated PyFluent
    features."""

    pass


class PyFluentUserWarning(UserWarning):
    """Provides the common warning class for warnings generated from user code."""

    pass


class FluentDevVersionWarning(PyFluentUserWarning):
    """Warning raised when a released PyFluent version is used with a development version of Fluent."""

    pass


# Deriving from the base Warning instead of any derived classes
# to have less chance of being ignored by user configurations.
class InsecureGrpcWarning(Warning):
    """Warning raised when gRPC connection is insecure."""

    pass


warnings.filterwarnings("always", category=InsecureGrpcWarning)


def warning_for_fluent_dev_version(version):
    """Provides warning if Fluent develop branch is used."""
    from ansys.fluent.core import FluentVersion, config

    if FluentVersion(version) > FluentVersion(config.fluent_release_version):
        warnings.warn(
            "⚠️ Warning: You are using PyFluent with an unreleased or development version of Fluent.\n"
            "Compatibility is not guaranteed, and unexpected behavior may occur. Please use a released "
            "version of Fluent that is officially supported by this version of PyFluent.",
            FluentDevVersionWarning,
        )


class WarningControl:
    """Class to control warnings in PyFluent."""

    def enable(self):
        """Enables all PyFluent warnings."""
        warnings.simplefilter("default", PyFluentDeprecationWarning)
        warnings.simplefilter("default", PyFluentUserWarning)

    def disable(self):
        """Disables all PyFluent warnings."""
        warnings.simplefilter("ignore", PyFluentDeprecationWarning)
        warnings.simplefilter("ignore", PyFluentUserWarning)


warning = WarningControl()
